
/*
	findclust.c - Cluster labeling functions
*/



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "dlgbox.h"
#include "imgbuf.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "application.h"
#include "functions.h"
#include "findclust.h"

/* Define the hard maximum of clusters and the debug level (0,1,2,3) */

#define MAXCLUST	32768
#define CLUST_DBG	2

/* For cluster counting: Number of threshold points */

#define	CPTS		50

int clustmax;				/* Number of clusters actually found */
int* clustlist;	/* Cluster list */
int* clustnum1;	/* Cluster list */
int* clustnum2;	/* Cluster list */
int* clustsize;	/* Cluster list */
int* clustsiz2;	/* Cluster list */
int max_no_of_clusters;		/* the maximum number of clusters we are interested in */
int errflg;


GtkWidget *clust_label8;
GtkWidget *clust_label9;

typedef struct 
{
	int r; int g; int b; 
	float t; 
	int neighb;
	int minsize, maxclust;
	int numsamples;
	int lowdens;
	int labelopts;
	int tabular;
	image_type* img;
}
clust_params;

clust_params params;


/* Returns if the image at (x,y,z) is nonzero, irrespective of image type */

char is_set (image_type* img, int x, int y, int z)
{
long idx;

	idx = x + img->xmax*(y + img->ymax*z);
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = img->data;
			if (p[idx]!=0) return 1;
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			if (p[idx]!=0) return 1;
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			if (!F_ISZERO(p[idx])) return 1;
			break;
		}
		case B_RGB:
		{
			rgbtriplet* p = img->data;
			if ((p[idx].red!=0) || (p[idx].green!=0) || (p[idx].blue!=0)) return 1;
			break;
		}
	}
	return 0;
}
		
			

#define P(X,Y)		(p[(X) + dest->xmax*(Y)])
#define P3(X,Y,Z)	(p[(X) + dest->xmax*( (Y) + dest->ymax*(Z))])

/**************************************************************************

 ##### 
#     #  #       #####
#     #  #       #    #
#     #  #       #    #
#     #  #       #    #
#     #  #       #    #
 #####   ######  #####

   #
  # #    #        ####    ####   #####      #     #####  #    #  #    #
 #   #   #       #    #  #    #  #    #     #       #    #    #  ##  ##
#     #  #       #       #    #  #    #     #       #    ######  # ## #
#######  #       #  ###  #    #  #####      #       #    #    #  #    #
#     #  #       #    #  #    #  #   #      #       #    #    #  #    #
#     #  ######   ####    ####   #    #     #       #    #    #  #    #


**************************************************************************/





/*
	This is pass1, which creates the pre-clustered image in dest,
	treating src as a binary (zero / nonzero) image.
	Allocating dest is the caller's responsibility.
	src may be discarded after this call.
	
	To keep these functions clean, the 2D and the general (3D)
	case are treated in separate functions.
	The special case 2D has some limitations. It always acts on
	the first slice, and it returns a 2D image even when the 
	source image was a 3D stack.
*/

void pass1_2d (image_type* src, image_type* dest, int slice, char f8neigh)
{
int x, y;
int clustnum;  
short* p;
long idx;

	g_assert (src->xmax == dest->xmax);
	g_assert (src->ymax == dest->ymax);
	g_assert (dest->imgtype == B_SHORT);

	dp (1, "Find clusters, pass 1:");

	p = dest->data;
	for (idx=0; idx<src->ymax*src->xmax; idx++) 
		p[idx]=0;

   	/* Special treatment for upper row & left column */

	clustnum=0;
   	if (is_set(src,0,0,slice)) { clustnum=1; p[0]=1; }
   	for (y=1; y<src->ymax; y++) if (is_set(src,0,y,slice) )     /* left column */
   	{
      	if (P(0,y-1)>0) 			/* already has a neighbor? */
			P(0,y)=P(0,y-1); 
		else 
			P(0,y)=++clustnum;		/* No, start a new cluster */
   	}

   	for (x=1; x<src->xmax; x++) if (is_set (src,x,0,slice))     /* Upper row */
   	{
      	if ( P(x-1,0)>0)  			/* already has a neighbor? */
			P(x,0)=P(x-1,0); 
		else 
			P(x,0)=++clustnum;		/* No, start a new cluster */
   	}

   	/* Now scan the entire image, checking the mirrored L
   	//
   	// c A d
   	// B ?   ? will be set to B if B>0, else to A if A>0;
   	//         c and d accordingly, if f8neigh = 1
   	*/
      
   	for (y=1; y<src->ymax; y++)
      	for (x=1; x<src->xmax; x++)
      	{
         	if (is_set (src,x,y,slice))     /* Assign a cluster# to this pt */
	 		{
	    		if (P(x-1,y)>0) P(x,y)=P(x-1,y);
	    		else if (P(x,y-1)>0) P(x,y)=P(x,y-1);
	    		else if ((f8neigh!=0) && (P(x-1,y-1)>0)) P(x,y)=P(x-1,y-1);
	    		else if ((f8neigh!=0) && (P(x+1,y-1)>0)) P(x,y)=P(x+1,y-1);
	    		else P(x,y) = ++clustnum;	/* start new cluster */
	 		}
	 
      	}

	/* 	dest now contains a pre-clustered image.
		It is possible that some connected areas
		have different numbers -> remedy in pass 2
	*/
	
	clustmax = clustnum;	/* Make this maximum public */
	dp (1, " found %d raw clusters\n", clustmax);
      
}


void pass1_3d (image_type* src, image_type* dest, char f8neigh)
{
int x, y, z;
int clustnum;  
short* p;
long idx;
long numelems;

	g_assert (src->xmax == dest->xmax);
	g_assert (src->ymax == dest->ymax);
	g_assert (src->zmax == dest->zmax);
	g_assert (dest->imgtype == B_SHORT);

	dp (1, "Find 3D clusters, pass 1:");

	p = dest->data;
	numelems = src->zmax * src->ymax * src->xmax;
	for (idx=0; idx<numelems; idx++) p[idx]=0;

   	/* Startup: Special treatment for upper row & left column */

	clustnum=0;
   	if (is_set(src,0,0,0)) { clustnum=1; p[0]=1; }
   	for (y=1; y<src->ymax; y++) if (is_set(src,0,y,0) )     /* left column */
   	{
      	if (P3(0,y-1,0)>0) 			/* already has a neighbor? */
			P3(0,y,0)=P3(0,y-1,0); 
		else 
			P3(0,y,0)=++clustnum;		/* No, start a new cluster */
   	}

   	for (x=1; x<src->xmax; x++) if (is_set (src,x,0,0))     /* Upper row */
   	{
      	if ( P3(x-1,0,0)>0)  			/* already has a neighbor? */
			P3(x,0,0)=P3(x-1,0,0); 
		else 
			P3(x,0,0)=++clustnum;		/* No, start a new cluster */
   	}
	
	/* Still starting up: Slice z=0 as 2D special case */
     
	for (y=1; y<src->ymax; y++)
		for (x=1; x<src->xmax; x++)
			if (is_set (src,x,y,0) )
			{
				if (P3(x-1,y,0)>0) P3(x,y,0)=P3(x-1,y,0);
				else if (P3(x,y-1,0)>0) P3(x,y,0)=P3(x,y-1,0);
				else P3(x,y,0) = ++clustnum;
			}
	 
   	/* Now scan the entire image, checking the mirrored L -- 3D case 
	   TODO: I think the 8 neighbor case is not yet clean */
      
	for (z=1; z<src->zmax; z++)
   		for (y=1; y<src->ymax; y++)
      		for (x=1; x<src->xmax; x++)
      		{
         		if (is_set (src,x,y,z))     /* Assign a cluster# to this pt */
	 			{
	    			if (P3(x-1,y,z)>0) P3(x,y,z)=P3(x-1,y,z);
	    			else if (P3(x,y-1,z)>0) P3(x,y,z)=P3(x,y-1,z);
	    			else if (P3(x,y,z-1)>0) P3(x,y,z)=P3(x,y,z-1);
	    			else if ((f8neigh!=0) && (P3(x-1,y-1,z-1)>0)) P3(x,y,z)=P3(x-1,y-1,z-1);
	    			else if ((f8neigh!=0) && (P3(x+1,y-1,z-1)>0)) P3(x,y,z)=P3(x+1,y-1,z-1);
	    			else if ((f8neigh!=0) && (P3(x-1,y-1,z  )>0)) P3(x,y,z)=P3(x-1,y-1,z);
	    			else if ((f8neigh!=0) && (P3(x+1,y-1,z  )>0)) P3(x,y,z)=P3(x+1,y-1,z);
	    			else P3(x,y,z) = ++clustnum;	/* start new cluster */
	 			}

      		}

	/* 	dest now contains a pre-clustered image.
		It is possible that some connected areas
		have different numbers -> remedy in pass 2
	*/
	
	clustmax = clustnum;	/* Make this maximum public */
	dp (1, " found %d raw 3D clusters\n", clustmax);
      
}



/*
	COMPACT takes care that any cluster found in PASS 1
	which consists of 2 or more numbers is unified and gets
	assigned only one number.
	
	The individual steps are:
	1. Generate a table with all cluster numbers. Clusters with
	   multiple numbers are stored as consecutive blocks
	2. Generation of a translation table
	   Table may be sorted by cluster size
	3. Renumbering of the clusters in the buffer.


	Tableins: if the new cluster A is neighbor of B, look up B in the table.
   	If B already exists, insert A right after B.
	Otherwise, append A - B - 00 to the table.
	Also check if the current block already contains A elsewhere


*/

char tableins (int clustA, int clustB)
{   
int i, j, k;

	if (clustmax>=max_no_of_clusters)
	{
		if (!(errflg &0x01)) dp (0, "Alert: Table full\n");
		errflg |= 1;
		return 1;	/* Can't add any more */
	}
   	i = 0;
   	while ((clustnum1[i]!=clustB) && (i < clustmax)) i++;
   	if (clustnum1[i]==clustB) /* Insert A as new word */
   	{
      	j=i; while ((clustnum1[j]!=0) && (j>0)) j--;
      	k=i; while ((clustnum1[k]!=0) && (k<clustmax)) k++;
      	while ((j<=k) && (clustnum1[j]!=clustA)) j++;
      	if (clustnum1[j]!=clustA)  /* Only if A does not yet exist */
      	{
         	for (j=clustmax; j>i; j--) clustnum1[j]=clustnum1[j-1];
         	clustmax++; clustnum1[i]=clustA;
	   		dp (3, "Inserting %d to join %d at %d\n", clustA, clustB, i);
      	}
   	}
  	else  /* append A and B */
   	{
   		dp (3, "Adding %d %d at %d\n", clustA, clustB, clustmax);
      	clustnum1[clustmax++]=clustA;
      	if (clustA!=clustB) clustnum1[clustmax++]=clustB;
      	clustnum1[clustmax++]=0;
   	}
	return 1;
}


/* Please note that we use the same P macro, but on a different entity dest.
   To keep it compact, this function handles both 2D and 3D cases.
   Large areas are independent of the image(s), anyway.
   The tables that are being generated by this function have the format
   aa - bb - cc - 0 - dd - ee - 0 - ...
   where the raw clusters aa, bb, cc are one contiguous cluster, and
   dd, ee another one.
   Please note that v serves as some kind of flag. Whenever the cluster
   has a forward neighbor, v is set to 1 through tableins(). If at the end
   of testing for neighbors v is still zero, it is an isolated cluster
   that gets a new entry in the tables.
*/
   

void compact (image_type* dest, int f8neigh)			/* this is pass 2 */
{
int x, y, z, i, j, c1, c2, a;
int sortrng,  min, minidx, iter;
short* p;
char v;

	p = dest->data;

   	dp (1, "Building tables...\n");

   	clustmax=0; clustnum1[0]=0;
   	if (f8neigh==0)
   	{
		for (z=0; z<dest->zmax; z++)
			for (y=0; y<dest->ymax-1; y++)
				for (x=0; x<dest->xmax-1; x++)
        	    	if (P3(x,y,z))
            		{
	       				if (P3(x+1,y,z)!=0) tableins (P3(x+1,y,z), P3(x,y,z));	/* Has a forward neighbor */
	   	       			if (P3(x,y+1,z)!=0) tableins (P3(x,y+1,z), P3(x,y,z));
	   	       			if ((y>0) && (P3(x,y-1,z)!=0)) tableins (P3(x,y-1,z), P3(x,y,z));
	   	       			if (z < dest->zmax-1)	/* This is special for the 3D case */
						{
							if (P3(x,y,z+1)!=0) tableins (P3(x,y,z+1), P3(x,y,z));
			       			if ((P3(x+1,y,z)==0) 
								&& (P3(x,y+1,z)==0)
								&& (P3(x,y,z+1)==0))
								tableins (P3(x,y,z), P3(x,y,z));  /* Isolated cluster - start a new one*/
						}
						else
			       			if ((P3(x+1,y,z)==0) && (P3(x,y+1,z)==0)) tableins (P3(x,y,z), P3(x,y,z));  /* Start a new cluster */
        	    	}
   	}
   	else
   	{
		for (z=0; z<dest->zmax; z++)
			for (y=0; y<dest->ymax-1; y++)
				for (x=0; x<dest->xmax-1; x++)
        	    	if (P3(x,y,z))
            	 	{
	       				v=0;
						if (P3(x+1,y,z)!=0) v=tableins (P3(x+1,y,z), P3(x,y,z));
   	       				if (P3(x,y+1,z)!=0) v=tableins (P3(x,y+1,z), P3(x,y,z));
		       			if (P3(x+1,y+1,z)!=0) v=tableins (P3(x+1,y+1,z), P3(x,y,z));
   		       			if ((x>0) && (P3(x-1,y+1,z)!=0)) v=tableins (P3(x-1,y+1,z), P3(x,y,z));

						if (z < dest->zmax-1)	/* This is special for the 3D case */
						{
		       				if (P3(x+1,y,z+1)!=0) v=tableins (P3(x+1,y,z+1), P3(x,y,z+1));
   		       				if (P3(x,y+1,z+1)!=0) v=tableins (P3(x,y+1,z+1), P3(x,y,z+1));
			       			if (P3(x+1,y+1,z+1)!=0) v=tableins (P3(x+1,y+1,z+1), P3(x,y,z+1));
   		    	   			if ((x>0) && (P3(x-1,y+1,z+1)!=0)) v=tableins (P3(x-1,y+1,z+1), P3(x,y,z+1));
							if (z>0)
							{
			       				if (P3(x+1,y,z-1)!=0) v=tableins (P3(x+1,y,z-1), P3(x,y,z-1));
   	    		   				if (P3(x,y+1,z-1)!=0) v=tableins (P3(x,y+1,z-1), P3(x,y,z-1));
		       					if (P3(x+1,y+1,z-1)!=0) v=tableins (P3(x+1,y+1,z-1), P3(x,y,z-1));
   		       					if ((x>0) && (P3(x-1,y+1,z-1)!=0)) v=tableins (P3(x-1,y+1,z), P3(x,y,z-1));
							}
							if (!v) tableins (P3(x,y,z),P3(x,y,z));
							
						}
						else
		    	   			if (!v) tableins (P3(x,y,z), P3(x,y,z));
            		}

   	}

   	/* Now generate the translation table. Clustnum1 is the raw cluster number,
	   clustnum2 is the reduced (compacted) cluster number */

   	j=1;
   	for (i=0; i<clustmax; i++)
   	{
      	clustnum2[i]=0;             /* Retain intermediate zeroes */
      	if (clustnum1[i]==0) j++;   /* New cluster number */
      	else clustnum2[i]=j;        /* Add to table */
   	}

	if (debuglevel>1) 
	{
		dp (2, "Cluster translation table before sorting:\n");
   		dp (2, "No.   Origin  Final\n");
	   	dp (2, "===================\n");
	   	for (i=0; i<=clustmax; i++) dp (2, "%5d  %5d  %4d\n", i, clustnum1[i], clustnum2[i]);
    }

	/* Sort translation table and remove the zeroes */

   	for (sortrng=0; sortrng<clustmax-1; sortrng++)
   	{
      	min = clustnum1[sortrng]; minidx=sortrng;
      	for (i=sortrng+1; i<clustmax; i++)
      	{
	 		if (min>clustnum1[i]) { min=clustnum1[i]; minidx=i; }
      	}
      	if (minidx!=sortrng)
      	{
	 		a=clustnum1[minidx]; clustnum1[minidx]=clustnum1[sortrng]; clustnum1[sortrng]=a;
	 		a=clustnum2[minidx]; clustnum2[minidx]=clustnum2[sortrng]; clustnum2[sortrng]=a;
      	}
   	}

   	i = 0; while ((i<clustmax) && (clustnum1[i]==0)) i++;
   	if (i>0)
   	{
      	for (j=0; j<clustmax-i+1; j++)
      	{
         	clustnum1[j]=clustnum1[j+i]; clustnum2[j]=clustnum2[j+i];
      	}
      	clustmax -= i;
   	}

   /*  We may still have "bridges" with the following morphology:
   //
   //   **   **
   //   **   **
   // ************
   // ************
   //
   // These need to be joined in a separate pass:
   */

	iter=1;
	while (iter)
	{
		i=0; iter=0;
   		while (i<clustmax-1)
   		{
    		if (clustnum1[i]==clustnum1[i+1]) 
      		{
	 			c1 = clustnum2[i+1];  c2 = clustnum2[i];	/* these are to be joined */
				dp (2, "Bridge removal: Joining %d and %d\n",c1,c2);
	 			for (j=0; j<clustmax; j++) if (clustnum2[j]==c1) clustnum2[j]=c2;
				for (j=i+1; j<clustmax; j++)
	 			{
	    			clustnum1[j-1]=clustnum1[j]; clustnum2[j-1]=clustnum2[j]; 
	 			}
	 			clustmax--;
				iter++;
      		}
      		else
         		i++;
   		}
		dp (2, "Another bridge iteration done. %d bridges removed\n",iter);
	}

	/* Renumber the cluster numbers to eliminate the gaps */

   	for (i=0; i<clustmax; i++) clustlist[i]=0;
   	for (i=0; i<clustmax; i++) clustlist[clustnum2[i]]=1;
   	j=0;
   	for (i=1; i<clustmax; i++) 
   	{
      	if (clustlist[i]!=0) j++;
      	clustlist[i]=j;
   	}
   	for (i=0; i<clustmax; i++) clustnum2[i]=clustlist[clustnum2[i]];


	if (debuglevel>1)
	{
	 	dp (2, "\n\nCluster translation table after sorting:\n");
   		dp (2, "No.   Origin  Final\n");
	   	dp (2, "===================\n");
   		for (i=0; i<=clustmax; i++) dp (2, "%5d  %5d  %4d\n", i, clustnum1[i], clustnum2[i]);
   	}
	dp (1, "Pass 2 finished.\n");

}

/*
	The third pass generates a consecutive, gap-free cluster numbering
	Furthermore, the cluster size will be determined and the resulting
	cluster list generated.
	This pass also takes care of the elimination of clusters that fall below
	the size limit.
*/	 

int translate (int clustA)   /* binary lookup */
{
int i, j, mid;
unsigned char found;

   	i=0; j=clustmax-1; found=0;
   	do
   	{
      	mid=(i+j)/2;
      	if (clustnum1[mid]>clustA) j=mid-1;
      	else if (clustnum1[mid]<clustA) i=mid+1;
      	else found=1;
   	} 
	while ((i<=j) && (found==0));
   	if (found!=0) return clustnum2[mid]; else return 0;
}


int compress (image_type* dest, cluster_type* list, int fMinclustsize)	/* This is pass 3 */
{
int x, y, z, i, j, a;
int buf, prev, prevt, tmpbuf;
int sortrng, max, maxidx;
int clustmax1, clustmax2;
short* p;

	p = dest->data;

   	for (i=0; i<max_no_of_clusters; i++) 
	{ 
		clustsize[i]=0; clustlist[i]=0; 
		if (list) 
		{
			list[i].size=0;
			list[i].cog_x=0; list[i].cog_y=0; list[i].cog_z=0;
			list[i].bbox_x1=-1; list[i].bbox_x2=0;
			list[i].bbox_y1=0; list[i].bbox_y2=0;
			list[i].bbox_z1=0; list[i].bbox_z2=0;
		}
		
	}
	
	/* Perform the translation pixel-by-pixel */

	prev=-1; clustmax1=0; prevt=0;
	for (z=0; z<dest->zmax; z++)
		for (y=0; y<dest->ymax; y++)
			for (x=0; x<dest->xmax; x++)
			{
				buf = P3(x,y,z);
				if (buf!=0)
				{
					if (buf==prev) buf = tmpbuf = prevt;
					else buf = translate (tmpbuf=buf);
					
					if (buf==0)
					{
						P3(x,y,z)=-1; 	/* Not found -- should never happen */
						if (!(errflg & 2)) dp (0, "Alert - Findclusters Pass 3: Illegal cluster number\n");
						errflg |= 2;
					}
					else 
					{   
						prev = P3(x,y,z); prevt=buf;
						P3(x,y,z)=buf; clustsize[buf]++; clustlist[buf]=1;
						if (clustmax1<buf) clustmax1=buf;
					}
				}
			}
	clustmax2=clustmax1;

	for (i=0; i<=clustmax1; i++) 
	if (clustsize[i]<fMinclustsize) 
	{
		clustlist[i]=0; clustsize[i]=0;
	}
		
	/* Before assigning the final cluster numbers, sort by size */

	for (i=0; i<=clustmax1; i++) clustsiz2[i]=0;
	j=0;

	dp (2, "\nIntermediate translation table:\n");

	for (i=0; i<=clustmax1; i++)
		if (clustlist[i]>0) 
		{ 
			dp (2, "%2d   %5d   %5d\n", j, clustsize[i], i);
			clustsiz2[j]=clustsize[i]; clustnum1[j]=i; 
			clustmax2=j+1; j++;
		}

	/* Sort translation table by clustsiz2 */

	for (sortrng=0; sortrng<clustmax2-1; sortrng++)
	{
		max = clustsiz2[sortrng]; maxidx=sortrng;
		for (i=sortrng+1; i<clustmax2; i++)
		{
			if (max<clustsiz2[i]) { max=clustsiz2[i]; maxidx=i; }
		}
		if (maxidx!=sortrng)
		{
			a=clustnum1[maxidx]; clustnum1[maxidx]=clustnum1[sortrng]; clustnum1[sortrng]=a;
			a=clustsiz2[maxidx]; clustsiz2[maxidx]=clustsiz2[sortrng]; clustsiz2[sortrng]=a;
		}
	}
	
	if (debuglevel>1)
	{
   		dp (2, "\nIntermediate translation table after sorting:\n");
   		for (i=0; i<clustmax2; i++)
			dp (2, "%2d   %5d   %5d\n", i, clustsiz2[i], clustnum1[i]);
   	}
   	for (i=0; i<clustmax2; i++) clustlist[clustnum1[i]]=i+1;

	if (debuglevel>1)
	{
		dp (2, "\n\nCluster translation table after pass 3:\n");
   		dp (2, "No.   Cluster  Size\n");
   		dp (2, "===================\n");
   		for (i=0; i<=clustmax1; i++) dp (2, "%5d  %4d %5d\n", i, clustlist[i], clustsize[i]);
	}

   	/* Convert the raw cluster numbers in BUF into the final cluster numbers
	   and determine size, bbox, center of gravity */
   
	for (z=0; z<dest->zmax; z++)
		for (y=0; y<dest->ymax; y++)
      		for (x=0; x<dest->xmax; x++)
      		{
         		if (P3(x,y,z)>0)  /* TODO: There are negative values, but shouldn't */
				{
			 		i = clustlist[P3(x,y,z)];
					P3(x,y,z) = i;
					if (list!=NULL)
					{
						list[i].size++;
						list[i].cog_x += x;
						list[i].cog_y += y;
						list[i].cog_z += z;
						if (list[i].bbox_x1 == -1)
						{
							list[i].bbox_x1 = list[i].bbox_x2 = x;
							list[i].bbox_y1 = list[i].bbox_y2 = y;
							list[i].bbox_z1 = list[i].bbox_z2 = z;
						}
						else
						{
							if (x < list[i].bbox_x1) list[i].bbox_x1 = x;
							if (y < list[i].bbox_y1) list[i].bbox_y1 = y;
							if (z < list[i].bbox_z1) list[i].bbox_z1 = z;
							if (x > list[i].bbox_x2) list[i].bbox_x2 = x;
							if (y > list[i].bbox_y2) list[i].bbox_y2 = y;
							if (z > list[i].bbox_z2) list[i].bbox_z2 = z;
						}
					}
				}
      		}
		
	/* Finish up the COG variables */
	
	if (list!=NULL)
	{
		dp (2, "Cluster  Size            COG                  BBox\n");

		for (i=0; i<clustmax2; i++)
		{
			if (list[i].size!=0)
			{
				list[i].cog_x /= list[i].size;
				list[i].cog_y /= list[i].size;
				list[i].cog_z /= list[i].size;
			}
			
			dp (2, "%4d   %6d    (%4li %4li %4li)   [(%4d %4d %4d) (%4d %4d %4d)]\n",
						i, list[i].size, list[i].cog_x, list[i].cog_y, list[i].cog_z,
						list[i].bbox_x1,list[i].bbox_y1, list[i].bbox_z1, 
						list[i].bbox_x2,list[i].bbox_y2, list[i].bbox_z2);
		}	
	} 
	
	dp (1, "Finished pass 3\n");
	return clustmax2;
}


/* Threshold an image (binarize) */

void clust_thresh (image_type* img, int r, int g, int b, float t, int thrtype)
{
long i,idx;


	idx = img->xmax * img->ymax * img->zmax;
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p;
			p = img->data;
			if (CLUST_DBG) fprintf (stderr, "Cluster: Thresholding CHAR image with t=%f\n",t);
			if (thrtype==0)
			{
				for (i=0; i<idx; i++)
					if ((float)p[i] >= t) p[i]=1; else p[i]=0;
			}
			else
			{
				for (i=0; i<idx; i++)
					if ((p[i]>0) && ((float)p[i] <= t)) p[i]=1; else p[i]=0;
			}
			img->minval.imin=0; img->maxval.imax=1;
			break;
		}
		case B_SHORT:
		{
			short* p;
			p = img->data;
			if (CLUST_DBG) fprintf (stderr, "Cluster: Thresholding SHORT image with t=%f\n",t);
			if (thrtype==0)
			{
				for (i=0; i<idx; i++)
					if ((float)p[i] >= t) p[i]=1; else p[i]=0;
			}
			else
			{
				for (i=0; i<idx; i++)
					if ((p[i]>0) && ((float)p[i] <= t)) p[i]=1; else p[i]=0;
			}
			img->minval.imin=0; img->maxval.imax=1;
			break;
		}
		case B_FLOAT:
		{
			float* p;
			p = img->data;
			if (CLUST_DBG) fprintf (stderr, "Cluster: Thresholding FLOAT image with t=%f\n",t);
			if (thrtype==0)
			{
				for (i=0; i<idx; i++)
					if (p[i] >= t) p[i]=1.0; else p[i]=0.0;
			}
			else
			{
				for (i=0; i<idx; i++)
					if ((p[i]>0) && (p[i] <= t)) p[i]=1.0; else p[i]=0.0;
			}
			img->minval.fmin=0.0; img->maxval.fmax=1.0;
			break;
		}
		case B_RGB:
		{
			rgbtriplet* p;
			p = img->data;
			if (CLUST_DBG) fprintf (stderr, "Cluster: Thresholding RGB image with t=(%d,%d,%d)\n",r,g,b);
			if (thrtype==0)
			{
				for (i=0; i<idx; i++)
				{
					if (p[i].red>=r) p[i].red=1; else p[i].red=0;
					if (p[i].green>=r) p[i].green=1; else p[i].green=0;
					if (p[i].blue>=r) p[i].blue=1; else p[i].blue=0;
				}
			}
			else
			{
				for (i=0; i<idx; i++)
				{
					if ((p[i].red>0) && (p[i].red<=r)) p[i].red=1; else p[i].red=0;
					if ((p[i].green>0) && (p[i].green<=r)) p[i].green=1; else p[i].green=0;
					if ((p[i].blue>0) && (p[i].blue<=r)) p[i].blue=1; else p[i].blue=0;
				}
			}
			img->minval.imin=0; img->maxval.imax=1;
			break;
		}
	}
	img->uid = next_uid++;
}




/**************************************************************************


#     #
##    #  ######  #    #
# #   #  #       #    #
#  #  #  #####   #    #
#   # #  #       # ## #
#    ##  #       ##  ##
#     #  ######  #    #

   #
  # #    #        ####    ####   #####      #     #####  #    #  #    #
 #   #   #       #    #  #    #  #    #     #       #    #    #  ##  ##
#     #  #       #       #    #  #    #     #       #    ######  # ## #
#######  #       #  ###  #    #  #####      #       #    #    #  #    #
#     #  #       #    #  #    #  #   #      #       #    #    #  #    #
#     #  ######   ####    ####   #    #     #       #    #    #  #    #



**************************************************************************/


/*
	The new algorithm is a one-pass version. It starts with a binary image
	with inverted signs (i.e. -1 signifies an unprocessed cluster).
	We run across the image, and whenever we encounter a -1 pixel, we
	floodfill that region, giving it a new unique number, and entering
	it in the cluster list. This cluster list is then postprocessed
	by size and renumbered.
*/


typedef struct
{
	int num;
	int srcnum;
	int bx1, bx2, by1, by2, bz1,bz2;
	int size; int boundsize;
	long Sx, Sy;
	float compactness;	/* Compactness of cluster */
	float aspect;		/* Aspect ratio of cluster */
	float irreg;		/* Irregularity (cv) of boundary */
}
clust_id_struct;

clust_id_struct clust_id;
clust_id_struct* clusttbl;	/* Store the initial cluster list */
int clust_flood_recursion_level;
int clust_flood_breakflg;


/* Note that the floodfill algorithm is recursive and thus stack-heavy.
   At this time we need to abort it at level 130,000 to prevent a segfault.
   To allow larger clusters, this algorithm needed to be replaced
   by the region growing algorithm below.
   However, at this time, we use a slightly updated and generalized version
   to perform interactive shape analysis (Fourier descriptors).
   This implies that clust_floodfill assumes a contiguous nonzero region.
   Upon exit, the filled cluster has the value 2, and the boundary is marked 
   with the value 3.
   Note that floodfill is a 2D operation only.
   Note that the image MUST be of type B_SHORT or floodfill fails.
   
   Both floodfill and region growing fill a contiguous area that requires
   the seed point (x,y) as its member with the value val.
*/


void clust_floodfill (image_type* img, clust_id_struct *clust, int x, int y, int f8ngh)
{

	if (clust_flood_recursion_level > 130000) clust_flood_breakflg=1;
	if (clust_flood_breakflg) return;

	/* Abort recursion if out of range */
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) return;
	clust_flood_recursion_level++;
	
	/* Only consider data points of value 1, i.e., points that are unvisited */
	
	if ( ((short*)img->data)[x + img->xmax*y] ==1)
	{
		/* Update the cluster info (bbox and size) */
		
		clust->size++;
		clust->bx1 = MIN(x,clust->bx1);
		clust->bx2 = MAX(x,clust->bx2);
		clust->by1 = MIN(y,clust->by1);
		clust->by2 = MAX(y,clust->by2);
		clust->Sx += x;
		clust->Sy += y;
		((short*)img->data)[x + img->xmax*y] ++;		/* Mark this pixel as visited */

		/* Check if this is a boundary pixel (add 1 if so). It is more convenient to
			determine boundary points here instead of later, i.e., after cluster outlining */

		if ((x==0) || (y==0) || (x==img->xmax-1) || (y==img->ymax-1))
		{
			((short*)img->data)[x + img->xmax*y] ++;		/* Mark this pixel as boundary */
			clust->boundsize++;
		}
		else												/* and exclude that we run ouside of the image in the next step */
		{
			if (	   (((short*)img->data)[x-1 + img->xmax*y] ==0)
					|| (((short*)img->data)[x+1 + img->xmax*y] ==0)
					|| (((short*)img->data)[x + img->xmax*(y-1)] ==0)
					|| (((short*)img->data)[x + img->xmax*(y+1)] ==0))
			{
				((short*)img->data)[x + img->xmax*y] ++;		/* Mark this pixel as boundary */
				clust->boundsize++;
			}
			else if (f8ngh)
			{
				if (	   (((short*)img->data)[x-1 + img->xmax*(y-1)] ==0)
					|| (((short*)img->data)[x+1 + img->xmax*(y-1)] ==0)
					|| (((short*)img->data)[x-1 + img->xmax*(y+1)] ==0)
					|| (((short*)img->data)[x+1 + img->xmax*(y+1)] ==0))
				{
					((short*)img->data)[x + img->xmax*y] ++;		/* Mark this pixel as boundary */
					clust->boundsize++;
				}
			}
		}

		/* and recurse to its four / eight neighbors */
		
		clust_floodfill (img, clust, x-1,y, f8ngh);
		clust_floodfill (img, clust, x+1,y, f8ngh);
		clust_floodfill (img, clust, x,y-1, f8ngh);
		clust_floodfill (img, clust, x,y+1, f8ngh);
		if (f8ngh)
		{
			clust_floodfill (img, clust, x-1,y-1, f8ngh);
			clust_floodfill (img, clust, x+1,y-1, f8ngh);
			clust_floodfill (img, clust, x-1,y+1, f8ngh);
			clust_floodfill (img, clust, x+1,y+1, f8ngh);
		}
	}
	clust_flood_recursion_level--;
}


/* Less efficient than floodfill, but unlimited */

void clust_regiongrow (image_type* img, int x0, int y0, int val, int f8ngh)
{
short *p;
int itercnt, changecnt, x,y, xm,ym, x1, x2, y1, y2;
long idx;


	g_assert (img->imgtype==B_SHORT);
	p = img->data;
	xm=img->xmax-1; ym=img->ymax-1;
	
	/* Initial scan limits around x0,y0 */
	
	x1 = MAX (0,x0-5);
	y1 = MAX (0,y0-5);
	x2 = MIN (xm,x0+5);
	y2 = MIN (ym,y0+5);
	
	
	itercnt=0; p[x0 + img->xmax*y0]=val;	/* set seed point */
	do
	{
		itercnt++; changecnt=0;
		dp (2, "Cluster region grow, value %d. Iteration %d: ", val, itercnt);
	
		/* With each ieration, we flip directions to speed up scanning */
		
		if (itercnt & 0x01)
		{
	
			for (y=y1; y<=y2; y++)
				for (x=x1; x<=x2; x++)
				{
					idx = x + img->xmax*y;
					if (f8ngh)
					{
						if ( (p[idx]==-1)
							&& (((x>0) && (p[idx-1]==val))
							|| ((x<xm) && (p[idx+1]==val))
							|| ((y>0) && (p[idx-img->xmax]==val))
							|| ((y<ym) && (p[idx+img->xmax]==val))

							|| ((x> 0) && (y> 0) && (p[idx-1-img->xmax]==val))
							|| ((x<xm) && (y> 0) && (p[idx+1-img->xmax]==val))
							|| ((x> 0) && (y<ym) && (p[idx-1+img->xmax]==val))
							|| ((x<xm) && (y<ym) && (p[idx+1+img->xmax]==val))   ))
						{
							p[idx] = val;
							changecnt++;

						   /* Update the cluster info (bbox and size) */

						   clust_id.size++;
						   clust_id.bx1 = MIN(x,clust_id.bx1);
						   clust_id.bx2 = MAX(x,clust_id.bx2);
						   clust_id.by1 = MIN(y,clust_id.by1);
						   clust_id.by2 = MAX(y,clust_id.by2);

						   /* Expand search area? */

						   if (x2 < x+5) x2 = MIN (x+5,xm);
						   if (x1 > x-5) x1 = MAX (x-5,0);
						   if (y2 < y+5) y2 = MIN (y+5,ym);
						   if (y1 > y-5) y1 = MAX (y-5,0);

						}
					}
					else
					{
						if ( (p[idx]==-1)
							&& (((x>0) && (p[idx-1]==val))
							|| ((x<img->xmax-1) && (p[idx+1]==val))
							|| ((y>0) && (p[idx-img->xmax]==val))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==val))))
						{
							p[idx] = val;
							changecnt++;

						   /* Update the cluster info (bbox and size) */

						   clust_id.size++;
						   clust_id.bx1 = MIN(x,clust_id.bx1);
						   clust_id.bx2 = MAX(x,clust_id.bx2);
						   clust_id.by1 = MIN(y,clust_id.by1);
						   clust_id.by2 = MAX(y,clust_id.by2);

						   /* Expand search area? */

						   if (x2 < x+5) x2 = MIN (x+5,xm);
						   if (x1 > x-5) x1 = MAX (x-5,0);
						   if (y2 < y+5) y2 = MIN (y+5,ym);
						   if (y1 > y-5) y1 = MAX (y-5,0);
						}

					}
				}
		}
		else
		{	
			for (y=y2; y>=y1; y--)
				for (x=x2; x>=x1; x--)
				{
					idx = x + img->xmax*y;
					if (f8ngh)
					{
						if ( (p[idx]==-1)
							&& (((x>0) && (p[idx-1]==val))
							|| ((x<xm) && (p[idx+1]==val))
							|| ((y>0) && (p[idx-img->xmax]==val))
							|| ((y<ym) && (p[idx+img->xmax]==val))

							|| ((x> 0) && (y> 0) && (p[idx-1-img->xmax]==val))
							|| ((x<xm) && (y> 0) && (p[idx+1-img->xmax]==val))
							|| ((x> 0) && (y<ym) && (p[idx-1+img->xmax]==val))
							|| ((x<xm) && (y<ym) && (p[idx+1+img->xmax]==val))   ))
						{
							p[idx] = val;
							changecnt++;

						   /* Update the cluster info (bbox and size) */

						   clust_id.size++;
						   clust_id.bx1 = MIN(x,clust_id.bx1);
						   clust_id.bx2 = MAX(x,clust_id.bx2);
						   clust_id.by1 = MIN(y,clust_id.by1);
						   clust_id.by2 = MAX(y,clust_id.by2);

						   /* Expand search area? */

						   if (x2 < x+5) x2 = MIN (x+5,xm);
						   if (x1 > x-5) x1 = MAX (x-5,0);
						   if (y2 < y+5) y2 = MIN (y+5,ym);
						   if (y1 > y-5) y1 = MAX (y-5,0);

						}
					}
					else
					{
						if ( (p[idx]==-1)
							&& (((x>0) && (p[idx-1]==val))
							|| ((x<img->xmax-1) && (p[idx+1]==val))
							|| ((y>0) && (p[idx-img->xmax]==val))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==val))))
						{
							p[idx] = val;
							changecnt++;

						   /* Update the cluster info (bbox and size) */

						   clust_id.size++;
						   clust_id.bx1 = MIN(x,clust_id.bx1);
						   clust_id.bx2 = MAX(x,clust_id.bx2);
						   clust_id.by1 = MIN(y,clust_id.by1);
						   clust_id.by2 = MAX(y,clust_id.by2);

						   /* Expand search area? */

						   if (x2 < x+5) x2 = MIN (x+5,xm);
						   if (x1 > x-5) x1 = MAX (x-5,0);
						   if (y2 < y+5) y2 = MIN (y+5,ym);
						   if (y1 > y-5) y1 = MAX (y-5,0);
						}

					}
				}
		}

		dp (2, "%d pixels\n",changecnt);

	}
	while (changecnt>0);
}



int new_pass1_2d (image_type* src, image_type* dest, int slice, char f8neigh, int minsize)
{
int x, y, i;
int clustnum;  
short* p;
long idx;
int sortrng, max, maxidx, buf;

	g_assert (src->xmax == dest->xmax);
	g_assert (src->ymax == dest->ymax);
	g_assert (dest->imgtype == B_SHORT);

	dp (1, "Find clusters, pass 1:");

	p = dest->data;
	for (idx=0; idx<src->ymax*src->xmax; idx++) 
		p[idx]=0;

   	
	/* Allocate clusttbl here, but we need to free it once everything is done, i.e. after the last pass */
	clusttbl = (clust_id_struct*) calloc (MAXCLUST, sizeof (clust_id_struct));
	if (!clusttbl) fatal_memerr();

	/* Prepare the initial thresholded image */
	
	for (y=0; y<src->ymax; y++)
      	for (x=0; x<src->xmax; x++)
      	{
         	if (is_set (src,x,y,slice))     /* Assign a cluster# to this pt */
	 		{
				p[x + dest->xmax * y] = -1;
			}
		}


	/* Run over the dest image seeking for -1 areas, floodfill them */

	clustnum=0;
	for (y=0; y<src->ymax; y++)
      	for (x=0; x<src->xmax; x++)
      	{
			if ((p[x + dest->xmax * y] == -1) && (clustnum < MAXCLUST))
			{
				clustnum++;
				clust_id.size=0;
				clust_id.bx1 = clust_id.bx2 = x;
				clust_id.by1 = clust_id.by2 = y;
				clust_id.num = clustnum;
				clust_flood_recursion_level=0; clust_flood_breakflg=0;
				clust_regiongrow (dest, x,y, clustnum, f8neigh);
				clusttbl[clustnum-1] = clust_id;
				clusttbl[clustnum-1].bz1 = clusttbl[clustnum-1].bz2 = 0;		/* Is a 2D algorithm :-( */
			}
		}

	/* 	dest now contains a pre-clustered image.
		It is possible that some clusters don't match 
		the size limit. They need to be deleted, and the 
		entire cluster set sorted by size.
	*/
	
	clustmax = clustnum;	/* Make this maximum public */
	dp (1, " found %d raw clusters\n", clustmax);
	
	dp (1, "Pass 2: Sorting...");					/* Sort the cluster list by size. */
	
	for (sortrng=0; sortrng<clustmax-1; sortrng++)
	{
		max = clusttbl[sortrng].size; maxidx=sortrng;
		for (i=sortrng+1; i<clustmax; i++)
		{
			if (max<clusttbl[i].size) { max=clusttbl[i].size; maxidx=i; }
		}
		if (maxidx!=sortrng)
		{
			clust_id = clusttbl[maxidx]; 
			clusttbl[maxidx]=clusttbl[sortrng];
			clusttbl[sortrng] = clust_id;
		}
	}
	dp (1, "(done) ...");
	
	/* Now we have a translation table. Each image value
		clusttbl[i].num must be translated into i+1 if its size
		is minsize or larger, or 0 (=deleted) if smaller. */
	
	/* TODO: Will be more efficient if we can use binary lookup */
	
	max=0;
	for (y=0; y<src->ymax; y++)
      	for (x=0; x<src->xmax; x++)
      	{
			if (p[x + dest->xmax * y]>0)
			{
				/* look it up */
				buf = p[x + dest->xmax * y];
				i=0; while ((i<clustmax) && (clusttbl[i].num!=buf)) i++;
				if (clusttbl[i].size >= minsize)
				{
					p[x + dest->xmax * y]=i+1;
					max = MAX(max,i+1);
				}
				else
					p[x + dest->xmax * y]=0;
			}
		}
					
	dp (1, "...OK. Clusters found: %d\n", max);
	return max;

}


/* This is the equivalent of the old algorithm's pass 3. It does
   cluster arrangement and postprocessing. Particularly,
   - it fills the output list
   - it computes center of gravity, aspect ratio etc

	Note about label options:
		Labelopt==0		rank by size (keep this for compatibility!)
		Labelopt==1		Pure binary
		Labelopt==2		Size (makes float image)
		Labelopt==3		compactness (makes float image)
		Labelopt==4		aspect ratio (makes float image)
		Labelopt==5		irregularity (makes float image)
   
   TODO: This is all 2D. We need separate algs for true 3D!
*/

char is_boundary_pixel (image_type* img, int x, int y, int z, int f8ngh)
{
long idx;
short* p;

	if ((x<=0) || (y<=0)) return 1;	/* Image border is always boundary */
	if ((x>=img->xmax-1) || (y>=img->ymax-1)) return 1;	/* Image border is always boundary */
	
	idx = x + img->xmax*(y + img->ymax*z); p = img->data;
	/* Check out the 4-neighborhood */
	if ((p[idx-1]==0) || (p[idx+1]==0)) return 1;
	if ((p[idx-img->xmax]==0) || (p[idx+img->xmax]==0)) return 1;
	if (f8ngh)	/* Check out the diagnoal neighborhood */
	{
		if ((p[idx-img->xmax-1]==0) || (p[idx+img->xmax-1]==0)) return 1;
		if ((p[idx-img->xmax+1]==0) || (p[idx+img->xmax+1]==0)) return 1;
	}
	return 0;
	
}


void new_pass2 (image_type* dest, int max, cluster_type* list, int f8ngh, int labelopts)
{
int i,x,y,z, size, bound;
long cogx,cogy,cogz,minrad,maxrad,rad,idx;
short* p;
float *fp; image_type fdest;					/* Floating-point return image */
int buf;
float Sx, Sxx;	/* for irregularity */

	g_assert (dest->imgtype == B_SHORT);		/* try to be clean */
	p=dest->data;

	dp (2, "Number  Value   Size   Cog_x  Cog_y Boundary Compact  Aspect Irregul%%.\n");

	for (i=0; i<max; i++)
	{

		/* Compute cluster by cluster within the bbox:
			1) Center of gravity
			2) Area (=size)
			3) Perimeter (shape: compactness = Perimeter_square/area)
			4) inner, outer radius (shape: aspect ratio = outer / inner circle
			5) CV of the boundary radii (irregularity)
		   Unfortunately, we need to go over the area twice.
		*/

		size=0; cogx=0; cogy=0; cogz=0;
		for (z=clusttbl[i].bz1; z<=clusttbl[i].bz2; z++)
			for (y=clusttbl[i].by1; y<=clusttbl[i].by2; y++)
				for (x=clusttbl[i].bx1; x<=clusttbl[i].bx2; x++)
				{
					if (p[x + dest->xmax*(y + dest->ymax*z)]==i+1)	/* then it's our cluster */
					{
						cogx+=x; cogy+=y; cogz+=z; size++;
					}
				}

		g_assert (size>0);	/* Software bug if not!! */

		cogx/=size; cogy/=size; cogz/=size;

		/* Now for the shape descriptors... */

		bound=0; Sx = 0; Sxx = 0;
		minrad=ARITHM(clusttbl[i].bx2-clusttbl[i].bx1,clusttbl[i].by2-clusttbl[i].by1);
		maxrad=0;
		for (z=clusttbl[i].bz1; z<=clusttbl[i].bz2; z++)
			for (y=clusttbl[i].by1; y<=clusttbl[i].by2; y++)
				for (x=clusttbl[i].bx1; x<=clusttbl[i].bx2; x++)
				{
					if ((p[x + dest->xmax*(y + dest->ymax*z)]==i+1)	/* then it's our cluster */
						&& (is_boundary_pixel (dest, x,y,z,f8ngh)))
					{
						bound++;	/* Count boundary pixels for perimeter - crude approximation since we count diagonals as 1 */
						rad = ARITHM(x-cogx,y-cogy);		/* SQUARED distance to centroid */
						minrad = MIN(minrad,rad);
						maxrad = MAX(rad,maxrad);
						Sx += sqrt(rad); Sxx += rad;
						if ((x==323) && (y==184))
						p[x + dest->xmax*(y + dest->ymax*z)]=250;
					}
				}

		clusttbl[i].boundsize = bound;
		clusttbl[i].compactness = (float)bound*bound/(float)size;
		if (minrad==0) minrad=1;
		clusttbl[i].aspect = sqrt(maxrad/minrad);
		clusttbl[i].irreg = SDEV(Sx, Sxx, bound) * bound / Sx;

		dp (2, "%6d %6d %6d  %5d %5d   %5d    %6.2f  %6.2f  %6.2f\n",
			i,i+1,size,cogx,cogy, bound, clusttbl[i].compactness, clusttbl[i].aspect, 100.0*clusttbl[i].irreg);

		if (list)
		{
			list[i].cog_x = cogx;
			list[i].cog_y = cogy;
			list[i].cog_z = cogz;
			list[i].size =  clusttbl[i].size;
			list[i].bbox_x1 =  clusttbl[i].bx1;
			list[i].bbox_x2 =  clusttbl[i].bx2;
			list[i].bbox_y1 =  clusttbl[i].by1;
			list[i].bbox_y2 =  clusttbl[i].by2;
			list[i].bbox_z1 = list[i].bbox_z2 = 0;	/* TODO: Needs to be mod'd for 3D */
			list[i].compactness = clusttbl[i].compactness;
			list[i].aspect = clusttbl[i].aspect;
			list[i].irreg = clusttbl[i].irreg;
		}
	}
	
	dp (2, "==================================================\n");

	/* If requested, label by any of the shape features
	   NOTE: We lose the link to the output list, of course. 
	   NOTE also: Any labelopts > 1 create a floating-point image and change dest. */
	
	if (labelopts<=1)			/* Labelopts==0 (default) is ranked by size. We've already done that */
	{							/* Labelopts==1 create a purely binary image, but B_SHORT for compatibility */

		if (labelopts==1)
		{
			for (z=0; z<dest->zmax; z++)
				for (y=0; y<dest->ymax; y++)
					for (x=0; x<dest->xmax; x++)		/* Gah! Ugly triple loop. Clean me UP!!! */
					{
						idx = x + dest->xmax*(y + dest->ymax*z);
						if (p[idx]>0) p[idx]=1;						/* Pure binary */
					}
		}
	}
	else							/* Labelopts > 1: Create a new FP image! */
	{
		if (allocate_image (&fdest, B_FLOAT, dest->xmax,dest->ymax,dest->zmax)==-1)
			fatal_memerr ();
		fp = fdest.data;

		for (z=0; z<dest->zmax; z++)
			for (y=0; y<dest->ymax; y++)
      			for (x=0; x<dest->xmax; x++)
      			{
					idx = x + dest->xmax*(y + dest->ymax*z);
					if (p[idx]>0)
					{
						/* look it up */
						buf = p[idx];
						switch (labelopts)
						{
							case 2: fp[idx] = (float)clusttbl[buf-1].size;
									break;
							case 3: fp[idx] = clusttbl[buf-1].compactness;
									break;
							case 4: fp[idx] = clusttbl[buf-1].aspect;
									break;
							case 5: fp[idx] = 100.0*clusttbl[buf-1].irreg;
									break;
						}
					}
					else
						fp[idx]=0;
				}

		freebuf (dest);
		handback (dest, &fdest);
	}

	/* Done with everything we need clusttbl for? Then free it */
	free (clusttbl);
	make_minmax (dest);


}


/*====================================================================
//
// The main control subroutine for the above code
//
//====================================================================*/


/*
	Findclust_main performs the mapping onto the image provided in img.
	img contains the binary source (the caller is responsible for
	preparing this from the source image), and upon exit, img contains
	the clusters. Upon exit, img's type is B_SHORT, limiting the
	algorithm to 32000 clusters.

	Parameters are img, the input and output image,
		maxclust, the cluster limit by the user
		f8neigh, whether to use 4-neighborhood (0) or 8-neighborhood
		minsize, the minimum cluster size (clusters below this size will be deleted)
		list, the list of clusters found

	If the calling subroutine passes a non-NULL pointer to list,
	findclust_main is supposed to return the cluster info via list
	up to a maximum of maxclust clusters.
	The function returns the number of clusters found.
*/


int findclust_2d (image_type* img, int slice, int maxclust, int f8neigh, int minsize, int labelopts, cluster_type* list)
{
image_type dest;
int clustmax2;

	errflg=0;
	dp (1, "Entering cluster finder algorithm\n");
	dp (1, "Max no. of clusters %d, min size %d, ",maxclust,minsize);
	if (f8neigh) dp (1, "8 neighbors"); else dp (1, "4 neighbors");
	dp (1, "\n");
		
	/* Allocate memory for the tables etc. */

	clustnum1 = (int*) calloc (maxclust, sizeof (int));
	clustsize = (int*) calloc (maxclust, sizeof (int));
	clustsiz2 = (int*) calloc (maxclust, sizeof (int));
	clustlist = (int*) calloc (maxclust, sizeof (int));
	clustnum2 = (int*) calloc (maxclust, sizeof (int));

	/* If any of these failed, exit */
	
	if ((!clustnum1) || (!clustsize) || (!clustsiz2) || (!clustlist) || (!clustnum2))
		return 0;

	/* Allocate the destination buffer */
      
	if (allocate_image (&dest, B_SHORT, img->xmax,img->ymax,1)==-1)
		fatal_memerr ();


	clustmax = 0;  max_no_of_clusters = maxclust;

	/* Call the cluster passes - new algorithm */
	
	clustmax2=new_pass1_2d (img, &dest, slice, f8neigh, minsize);
	new_pass2 (&dest, clustmax2, list, f8neigh, labelopts);

	if (errflg) dp (1, "Clusters: Error(s) occurred.\n");

	/* After we are done, free the buffer memory */
	
	free (clustnum1);
	free (clustnum2);
	free (clustsize);
	free (clustsiz2);
	free (clustlist);
	handback (img, &dest);
	
	return clustmax2;

}


/* TODO: findclust_3d uses the old algorithm. Needs to be fixed. Consequently, labelopts are only rudimentary. */


int findclust_3d (image_type* img, int maxclust, int f8neigh, int minsize, int labelopts, cluster_type* list)
{
image_type dest;
short *p;
long idx;
int x,y,z;
int clustmax2;

	errflg=0;
	dp (1, "Entering cluster finder algorithm\n");
	dp (1, "Max no. of clusters %d, min size %d, ",maxclust,minsize);
	if (f8neigh) dp (1, "8 neighbors"); else dp (1, "4 neighbors");
	dp (1, "\n");
		
	/* Allocate memory for the tables etc. */

	clustnum1 = (int*) calloc (maxclust, sizeof (int));
	clustsize = (int*) calloc (maxclust, sizeof (int));
	clustsiz2 = (int*) calloc (maxclust, sizeof (int));
	clustlist = (int*) calloc (maxclust, sizeof (int));
	clustnum2 = (int*) calloc (maxclust, sizeof (int));

	/* If any of these failed, exit */
	
	if ((!clustnum1) || (!clustsize) || (!clustsiz2) || (!clustlist) || (!clustnum2))
		return 0;

	/* Allocate the destination buffer */
      
	if (allocate_image (&dest, B_SHORT, img->xmax,img->ymax,img->zmax)==-1)
		fatal_memerr ();


	clustmax = 0;  max_no_of_clusters = maxclust;

	/* Call the cluster passes */
	
	pass1_3d (img, &dest, f8neigh);
	compact (&dest, f8neigh);
	clustmax2=compress (&dest,list,minsize);

	/* If labelopts==1 (pure binary), do it. */

	if (labelopts==1)
	{
		g_assert (dest.imgtype==B_SHORT);
		p = dest.data;
		for (z=0; z<dest.zmax; z++)
			for (y=0; y<dest.ymax; y++)
				for (x=0; x<dest.xmax; x++)		/* Gah! Ugly triple loop. Clean me UP!!! */
				{
					idx = x + dest.xmax*(y + dest.ymax*z);
					if (p[idx]>0) p[idx]=1;						/* Pure binary */
				}

	}

	make_minmax (&dest);
	if (errflg) dp (1, "Clusters: Error(s) occurred.\n");

	/* After we are done, free the buffer memory */
	
	free (clustnum1);
	free (clustnum2);
	free (clustsize);
	free (clustsiz2);
	free (clustlist);
	handback (img, &dest);
	
	return clustmax2;

}


/* This function uses the original image _and_ the clusterized image to add first-order
	intensity statistics to the cluster list. Works for both 2D and 3D.
	
	Inputs are:
		img				The original (unclustered) image
		clusterimg		The clusterized image, intepreted as binary (zero / nonzero)
		list			The cluster statistics list
		clustnum		The number of clusters
*/


void clust_measure_intens (image_type *img, image_type *clusterimg, cluster_type *list, int clustnum)
{
int i,x,y,z;
float buf;
short ibuf;


	/* Reset mean and stdev, because we'll use them as accumulators */
	
	for (i=0; i<clustnum; i++)
	{
		list[i].size=0;
		list[i].mean = list[i].stdev = 0.0;
	}

	/* Run over the entire image (everything where cluster image is nonzero)
		and collect intensities */

	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				readbuf (clusterimg, &ibuf, x,y,z);		/* Nonzero element? Get the cluster number for this pixel */
				if (ibuf)
				{
					buf = readbuf_flt (img, x,y,z);		/* Get intensity value */
					list[ibuf-1].mean +=buf;				/* collect Sx and Sxx */
					list[ibuf-1].stdev += SQR(buf);
					list[ibuf-1].size++;
				}
			}

	/* Since we already know list[i].size, compute mean and stdev for each list entry */

	dp (2, "Number  Value   Size   Sx    Sxx    Mean    Stdev.\n");

	for (i=0; i<clustnum; i++)
	{
		dp (2, "%6d %6d %6d      %8.2f  %8.2f", i,i+1, list[i].size, list[i].mean, list[i].stdev);

		buf = SDEV (list[i].mean, list[i].stdev, list[i].size);		/* Sx, Sxx, and N */
		list[i].mean /= list[i].size;
		list[i].stdev = buf;
		dp (2, "   %6.2f  %6.2f\n", list[i].mean, list[i].stdev);

	}
	dp (2, "==================================================\n");

}






void clust_report_tabular (cluster_type* list, int clustnum, int is_3d, int ext_rpt)
{
int i;
char s[255];
char* rpt;

	rpt = calloc (90*(clustnum+1), sizeof(clustnum));	/* Allow for 90 chars per line plus title */
	if (rpt==NULL) fatal_memerr();

	if (is_3d)
		sprintf (rpt, " Value   Size    CentX CentY CentZ");
	else
		sprintf (rpt, " Value   Size    CentX CentY   Compct  Aspect   Irreg");

	if (ext_rpt)
	{
		sprintf (s, "   Mean   Std.Dev");
		strcat (rpt, s);
	}

	strcat (rpt, "\n");

	
	for (i=0; i<clustnum; i++)
	{

		if (is_3d)
			sprintf (s, "%5d %7d  %5li %5li %5li", i+1,list[i].size,list[i].cog_x,list[i].cog_y, list[i].cog_z);
		else
			sprintf (s, "%5d %7d  %5li %5li", i+1,list[i].size,list[i].cog_x,list[i].cog_y);
		strcat (rpt, s);

		if (!is_3d)
		{
			sprintf (s, "    %6.2f  %6.2f  %6.2f", list[i].compactness, list[i].aspect, list[i].irreg);
			strcat (rpt, s);
		}

		if (ext_rpt)
		{
			sprintf (s, "    %6.2f  %6.2f", list[i].mean, list[i].stdev);
			strcat (rpt, s);
		}
		strcat (rpt, "\n");

	}
	ShowLongMessage (rpt, "Cluster Labeling Results");

}







/* Cluster counting: Number of clusters in dependency of the threshold.
   A similar function is used below in the GUI, and this one is
   a standalone version to be used in the macro interpreter.
   This function returns the maximum number of clusters found
   and sets the corresponding threshold in user[0] */



int clust_clustercount (image_type* img, float thresh, int numsamples, int minsize, int maxclust, int lowdens)
{
float delta, t, t1, t2, maxt;
int i,j, maxc;
image_type scratch;


	if (img->imgtype == B_FLOAT)
	{
		t1 = thresh;
		t2 = img->maxval.fmax;
	}
	else
	{
		t1 = thresh;
		t2 = img->maxval.imax;
	}
	delta = (t2-t1)/numsamples;

	t = t1; i=0; maxc=0; maxt=t;
	while (t<=t2)
	{

		clone_image (img, &scratch);
		clust_thresh (&scratch, (int)t, (int)t, (int)t, t, lowdens);

		if (img->zmax==1)
			j=findclust_2d (&scratch, 0, maxclust, 1 /*8*/, minsize, 0, NULL);
		else
			j=findclust_3d (&scratch, maxclust, 1, minsize, 0, NULL);

		freebuf (&scratch);

		if (j > maxc)
		{
			maxc = j; maxt = t;
		}

		dp (1, "Cluster counting: Point %d, threshold %f, clusters: %d\n",i,t,j);
		t += delta; i++;
	}
	
	mac_user[0] = maxt;		/* Return thershold through user[] arrax y */
	freebuf (&scratch);
	return maxc;			/* Return number of clusters */
}



/************************************************************

	Findclust-GUI
	
*************************************************************/




void on_clust_r_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.r = (int)(0.5+adj->value);
}

void on_clust_g_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.g = (int)(0.5+adj->value);
}

void on_clust_b_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.b = (int)(0.5+adj->value);
}

void on_clust_t_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.t = adj->value;
}

void on_clust_minsize_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.minsize = adj->value;
}

void on_clust_maxclust_adj_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.maxclust = adj->value;
}

void on_clust_numsamples_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	params.numsamples = adj->value;
}


void on_clust_f8button_button_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	params.neighb = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	if (params.neighb)
		gtk_label_set_text (GTK_LABEL(clust_label8), "8 neighbors");
	else
		gtk_label_set_text (GTK_LABEL(clust_label8), "4 neighbors");

}


void on_clust_lowdens_button_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	params.lowdens = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	if (params.lowdens)
		gtk_label_set_text (GTK_LABEL(clust_label9), "Low density");
	else
		gtk_label_set_text (GTK_LABEL(clust_label9), "Normal thresholding");

}

void on_clust_tabular_button_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	params.tabular = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Tabular output set to %d\n", params.tabular);

}


void on_clust_cancel_clicked ( GtkButton * button, gpointer data )
{
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_clust_clusterize ( GtkButton * button, gpointer data )
{
cluster_type* list;
int num;

	if (params.img == &mainimg) undo_nexus (params.img);

	list = NULL;
	if (params.tabular)
	{
		list = calloc (params.maxclust+10,sizeof(cluster_type));
		if (list==NULL) fatal_memerr();
	}

	clust_thresh (params.img, params.r, params.g, params.b, params.t, 0);

	if (params.img->zmax==1)
		num=findclust_2d (params.img, 0, params.maxclust, params.neighb, params.minsize, params.labelopts, list);
	else
		num=findclust_3d (params.img, params.maxclust, params.neighb, params.minsize, params.labelopts, list);

	if (params.img == &mainimg)
	{
		make_imlib_buffer_from_main_image (0); cur_slice=0;
		redraw (1);
	}
	
	if (params.tabular) clust_report_tabular (list, num, (params.img->zmax>1), 0);
	
	if (list) free (list);

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_clust_measure ( GtkButton *button, gpointer data )
{
cluster_type* list;
image_type clusterimg;
int num;

	list = NULL;
	if (params.tabular)			/* Which should always be TRUE */
	{
		list = calloc (params.maxclust+10,sizeof(cluster_type));
		if (list==NULL) fatal_memerr();
	}

	clone_image (params.img, &clusterimg);
	clust_thresh (&clusterimg, params.r, params.g, params.b, params.t, 0);

	if (clusterimg.zmax==1)
		num=findclust_2d (&clusterimg, 0, params.maxclust, params.neighb, params.minsize, params.labelopts, list);
	else
		num=findclust_3d (&clusterimg, params.maxclust, params.neighb, params.minsize, params.labelopts, list);

	clust_measure_intens (params.img, &clusterimg, list, num);
	
	if (params.tabular) clust_report_tabular (list, num, (params.img->zmax>1), 1);
	
	if (list) free (list);
	freebuf (&clusterimg);

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}



void on_clust_clustercount ( GtkButton * button, gpointer data )
{
float xdata[CPTS+1], ydata[CPTS+1];
float delta, t, t1, t2, maxt;
int i,j, maxc;
image_type scratch;

	if (params.img == &mainimg) undo_nexus (params.img);

	if (params.img->imgtype == B_RGB)
	{
		t1 = params.r; 
		if (params.g<t1) t1 = params.g;
		if (params.b<t1) t1 = params.b;
		t2 = params.img->maxval.imax;
	}
	else if (params.img->imgtype == B_FLOAT)
	{
		t1 = params.t;
		t2 = params.img->maxval.fmax;
	}
	else
	{
		t1 = params.t;
		t2 = params.img->maxval.imax;
	}
	delta = (t2-t1)/params.numsamples;

	t = t1; i=0; maxc=0; maxt=t;
	while ((t<=t2) && (i<CPTS))
	{

		clone_image (params.img, &scratch);
		clust_thresh (&scratch, (int)t, (int)t, (int)t, t, params.lowdens);

		if (params.img->zmax==1)
			j=findclust_2d (&scratch, 0, params.maxclust, params.neighb, params.minsize, params.labelopts, NULL);
		else
			j=findclust_3d (&scratch, params.maxclust, params.neighb, params.minsize, params.labelopts, NULL);

		freebuf (&scratch);

		if (j > maxc)
		{
			maxc = j; maxt = t;
		}

		dp (1, "Cluster counting: Point %d, threshold %f, clusters: %d\n",i,t,j);
		xdata[i] = t;
		ydata[i] = (float)j;
		t += delta; i++;
	}
	
	/* Now clusterize the main image at the cluster maximum */
	

	dp (1, "Cluster counting: Reclustering image at t=%f to obtain %d clusters\n",maxt,maxc);

	clust_thresh (params.img, (int)maxt, (int)maxt, (int)maxt, maxt, params.lowdens);
	if (params.img->zmax==1)
		j=findclust_2d (params.img, 0, params.maxclust, params.neighb, params.minsize, params.labelopts, NULL);
	else
		j=findclust_3d (params.img, params.maxclust, params.neighb, params.minsize, params.labelopts, NULL);

	if (params.img == &mainimg)
	{
		make_imlib_buffer_from_main_image (0); cur_slice=0;
		redraw (1);
	}

	/* And display the n_c (t) function */	
	
	graphme (xdata, ydata, i);

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}



void cclust_set_lbl_menu (GtkWidget *widget, gpointer user_data)
{
	params.labelopts = *(int*) user_data;
	dp (2, "Lableing options set to %d\n", params.labelopts);
}



void create_clustfind_wnd (image_type* img, int function)
{
  GtkWidget *clustfind_wnd;
  GtkWidget *clust_table1;
  GtkWidget *clust_label1, *clust_label2, *clust_label3, *clust_label4;
  GtkWidget *clust_label5, *clust_label6, *clust_label7, *clust_label10;
  GtkWidget *clust_okbutton;
  GtkWidget *clust_cancelbutton;
  GtkObject *clust_r_adj;
  GtkWidget *clust_r;
  GtkObject *clust_g_adj;
  GtkWidget *clust_g;
  GtkObject *clust_b_adj;
  GtkWidget *clust_b;
  GtkObject *clust_t_adj;
  GtkWidget *clust_t;
  GtkObject *clust_minsize_adj;
  GtkWidget *clust_minsize;
  GtkObject *clust_maxclust_adj;
  GtkWidget *clust_maxclust;
  GtkWidget *clust_f8button;
  GtkWidget *clust_tblbutton;
  GtkWidget* clust_lowdensbutton;
  GtkObject* clust_numsamples_adj;
  GtkWidget* clust_numsamples;
  GtkWidget *cclust_label_menu;
  GtkWidget *cclust_label_menu_menu;
  GtkWidget *glade_menuitem;
  
  float min,max;
  int ty;
  
  	params.labelopts=0;
	params.tabular = 0;
	imgminmax (img, &min, &max);  


	/* The functions we can handle:
		0: Regular cluster labeling, numbers clusters according to spec 
		1: Cluster count as a function of threshold C(T)
		2: Cluster measurement. Output tabular result only */

	
	clustfind_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (clustfind_wnd), "clustfind_wnd", clustfind_wnd);
	if (function==1)
		gtk_window_set_title (GTK_WINDOW (clustfind_wnd), "Find Clusters");
	else
		gtk_window_set_title (GTK_WINDOW (clustfind_wnd), "Cluster Counting");
	gtk_window_set_modal (GTK_WINDOW (clustfind_wnd), TRUE);

	/* The table is either 5x2 for grayscale or 7x2 for RGB.
	   Add two more controls if cluster in dependency of threshold */

	ty=5; if (img->imgtype == B_RGB) ty=7;
	if (function==0) ty +=2;		/* Labeling options */
	if (function==1) ty +=2;		/* Threshold and number samples */

	clust_table1 = gtk_table_new (ty, 2, FALSE);
	gtk_widget_ref (clust_table1);
	gtk_object_set_data_full (GTK_OBJECT (clustfind_wnd), "clust_table1", clust_table1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (clust_table1);
	gtk_container_add (GTK_CONTAINER (clustfind_wnd), clust_table1);

	/* Attach either gray value or rgb threshold selectors */

	if (img->imgtype == B_RGB)
	{	

		clust_label2 = MakeLabel (clustfind_wnd, clust_table1, "Red Threshold", "clust_label2", 0,0);
		clust_label3 = MakeLabel (clustfind_wnd, clust_table1, "Green Threshold", "clust_label3", 0,1);
		clust_label4 = MakeLabel (clustfind_wnd, clust_table1, "Blue Threshold", "clust_label4", 0,2);

		clust_r_adj = gtk_adjustment_new (min+1, min, max, 1, 10, 0);
		clust_r = MakeSpinbutton (GTK_ADJUSTMENT(clust_r_adj), clustfind_wnd, clust_table1, "clust_r", 1,0);
		gtk_signal_connect (GTK_OBJECT (clust_r), "value-changed", GTK_SIGNAL_FUNC (on_clust_r_adj_changed),clust_r_adj);

		clust_g_adj = gtk_adjustment_new (min+1, min, max, 1, 10, 0);
		clust_g = MakeSpinbutton (GTK_ADJUSTMENT(clust_g_adj), clustfind_wnd, clust_table1, "clust_g", 1,1);
		gtk_signal_connect (GTK_OBJECT (clust_g), "value-changed", GTK_SIGNAL_FUNC (on_clust_g_adj_changed),clust_g_adj);

		clust_b_adj = gtk_adjustment_new (min+1, min, max, 1, 10, 0);
		clust_b = MakeSpinbutton (GTK_ADJUSTMENT(clust_b_adj), clustfind_wnd, clust_table1, "clust_b", 1,2);
		gtk_signal_connect (GTK_OBJECT (clust_b), "value-changed", GTK_SIGNAL_FUNC (on_clust_b_adj_changed),clust_b_adj);

		params.r = params.g = params.b = min+1;

		ty=3;
	}
	else
	{
		clust_label2 = MakeLabel (clustfind_wnd, clust_table1, "Threshold", "clust_label2", 0,0);

		clust_t_adj = gtk_adjustment_new (min+1, min, max, 1, 10, 0);
		clust_t = MakeSpinbutton (GTK_ADJUSTMENT(clust_t_adj), clustfind_wnd, clust_table1, "clust_t", 1,0);
		gtk_signal_connect (GTK_OBJECT (clust_t), "value-changed", GTK_SIGNAL_FUNC (on_clust_t_adj_changed),clust_t_adj);
		params.t = min+1;
		ty=1;


	}

	/* Add the controls for 4/8 neighbors, max number of clusters,min size */


	clust_label6 = MakeLabel (clustfind_wnd, clust_table1, "Max # of clusters", "clust_label6", 0,ty);
	clust_maxclust_adj = gtk_adjustment_new (MAXCLUST, 1, MAXCLUST, 1, 10, 0);
	clust_maxclust = MakeSpinbutton (GTK_ADJUSTMENT(clust_maxclust_adj), clustfind_wnd, clust_table1, "clust_maxclust", 1,ty);
	gtk_signal_connect (GTK_OBJECT (clust_maxclust), "value-changed", GTK_SIGNAL_FUNC (on_clust_maxclust_adj_changed),clust_maxclust_adj);
	params.maxclust = MAXCLUST;
	ty++;
	
	clust_label7 = MakeLabel (clustfind_wnd, clust_table1, "Min size", "clust_label7", 0,ty);
	clust_minsize_adj = gtk_adjustment_new (3,1,1000, 1, 10, 0);
	clust_minsize = MakeSpinbutton (GTK_ADJUSTMENT(clust_minsize_adj), clustfind_wnd, clust_table1, "clust_minsize", 1,ty);
	gtk_signal_connect (GTK_OBJECT (clust_minsize), "value-changed", GTK_SIGNAL_FUNC (on_clust_minsize_adj_changed),clust_minsize_adj);
	params.minsize = 3;
	ty++;

	clust_label8 = MakeLabel (clustfind_wnd, clust_table1, "4 neighbors", "clust_label8", 0,ty);
	clust_f8button = Togglebutton_in_table (clustfind_wnd,clust_table1,"4/8 toggle","clust_f8button",1,ty,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (clust_f8button), FALSE);
	gtk_signal_connect (GTK_OBJECT (clust_f8button), "toggled",
                      GTK_SIGNAL_FUNC (on_clust_f8button_button_toggled),clust_f8button);
	params.neighb = 0;
	ty++;


	if (function==0)		/* Add the controls for different labeling functions */
	{
		clust_label10 = MakeLabel (clustfind_wnd, clust_table1, "Label clusters by", "clust_label10", 0,ty);
		cclust_label_menu = Make_Optionmenu (clustfind_wnd,clust_table1,"cclust_label_menu",1,ty);
		cclust_label_menu_menu = gtk_menu_new ();
		glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "Size Rank", (gpointer) &menuopt0);
		glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "None (pure binary)", (gpointer) &menuopt1);

		if (img->zmax==1)			/* Only for 2D images as long as 3D uses the old algorithm */
		{
			glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "Size", (gpointer) &menuopt2);
			glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "Compactness", (gpointer) &menuopt3);
			glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "Aspect ratio", (gpointer) &menuopt4);
			glade_menuitem = Option_menuitem (cclust_label_menu_menu, GTK_SIGNAL_FUNC(cclust_set_lbl_menu), "Irregularity", (gpointer) &menuopt5);		
		}

		gtk_option_menu_set_menu (GTK_OPTION_MENU (cclust_label_menu), cclust_label_menu_menu);
		gtk_option_menu_set_history (GTK_OPTION_MENU (cclust_label_menu), 1);
		params.labelopts=1;		/* Match the default option */
		ty++;

		clust_label10 = MakeLabel (clustfind_wnd, clust_table1, "Output", "clust_label10", 0,ty);
		clust_tblbutton = Checkbutton_in_table (clustfind_wnd,clust_table1,"Tabular results","clust_tabbtn",1,ty,NULL);
		gtk_signal_connect (GTK_OBJECT (clust_tblbutton), "toggled",
                    	GTK_SIGNAL_FUNC (on_clust_tabular_button_clicked), clust_tblbutton);
		ty++;
	}
	else if (function==1)		/* add the controls for clusters in dependency of threshold */
	{
		clust_label9 = MakeLabel (clustfind_wnd, clust_table1, "Normal Thresholding", "clust_label9", 0,ty);
		clust_lowdensbutton = Togglebutton_in_table (clustfind_wnd,clust_table1,"Thresholding toggle","clust_lowdensbutton",1,ty,NULL);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (clust_lowdensbutton), FALSE);
		gtk_signal_connect (GTK_OBJECT (clust_lowdensbutton), "toggled",
	                      GTK_SIGNAL_FUNC (on_clust_lowdens_button_toggled),clust_lowdensbutton);
		params.lowdens = 0;
		ty++;
	
		clust_label10 = MakeLabel (clustfind_wnd, clust_table1, "Number of threshold samples", "clust_label10", 0,ty);
		clust_numsamples_adj = gtk_adjustment_new (20, 2, CPTS, 1, 10, 10);
		clust_numsamples = MakeSpinbutton (GTK_ADJUSTMENT(clust_numsamples_adj), clustfind_wnd, clust_table1, "clust_numsamples", 1,ty);
		gtk_signal_connect (GTK_OBJECT (clust_numsamples), "value-changed", GTK_SIGNAL_FUNC (on_clust_numsamples_changed),clust_numsamples_adj);
		params.numsamples = 20;
		
		ty++;
	}
	else if (function==2)		/* Special handling of the measurement function */
	{
		params.labelopts=0;		/* Doesn't matter by what we'd label it -- we need a mask */
		params.tabular = 1;		/* Always with tabular output */
	}


	clust_okbutton = Button_in_table (clustfind_wnd, clust_table1, "OK", "clust_okbutton", 0,ty, NULL);
	clust_cancelbutton = Button_in_table (clustfind_wnd, clust_table1, "Cancel", "clust_okbutton", 1,ty, NULL);

	if (function==0)					/* regular clusterizing */
		gtk_signal_connect (GTK_OBJECT (clust_okbutton), "clicked",  GTK_SIGNAL_FUNC (on_clust_clusterize), clustfind_wnd);
	else if (function==1)				/* Clusters in dep of threshold */
		gtk_signal_connect (GTK_OBJECT (clust_okbutton), "clicked", GTK_SIGNAL_FUNC (on_clust_clustercount), clustfind_wnd);
	else if (function==2)				/* Cluster measurement */
		gtk_signal_connect (GTK_OBJECT (clust_okbutton), "clicked", GTK_SIGNAL_FUNC (on_clust_measure), clustfind_wnd);

	gtk_signal_connect (GTK_OBJECT (clust_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_clust_cancel_clicked), clustfind_wnd);
	gtk_signal_connect (GTK_OBJECT (clustfind_wnd), "destroy", GTK_SIGNAL_FUNC (on_clust_cancel_clicked), clustfind_wnd);


	gtk_widget_show (clustfind_wnd);
  
}

/* Driver for cluster algorithm: Clusterize present image */

void clustfind ()
{
	params.img = &mainimg;
	create_clustfind_wnd (params.img, 0);
}


/* Driver for cluster counting:
   Determine number of clusters in dependency of thresold */


void clustcnt ()
{
	params.img = &mainimg;
	create_clustfind_wnd (params.img, 1);
}


/* Driver for the Measure -> Cluster function. This function includes	
	grayscale metrics of the clusters and always outputs tabular results */

void clust_meas ()
{

	params.img = &mainimg;
	create_clustfind_wnd (params.img, 2);

}


/*********************************************************************

Other cluster-related functions


**********************************************************************/


/* Trace the boundary contour. Requires a B_SHORT image.
	Assumption: xs,ys is a valid start point. Boundary pixels have the value 3,
	and interior pixels have the value 2 (this is not important).
	Assumption: The contour does not touch the image boundary.
	Trick is we choose the pixel with the best connectivity to background
	with additional constraints that the path gets pushed away from visited
	pixels and prefers keeping its direction.

	Note that this is not a good implementation of a contour tracer.
	Better algorithms exist, and they should be at least able to backtrack
	if the algorithm gets stuck.

	For the purpose of the book it is sufficient, though.
*/

int runalong (image_type *img, int xs, int ys, int xc, int yc, float *ydata)
{
int cost[3][3];
int abort, x,y,xp,yp,n,i,j,k,i1,j1;
short buf;

	dp (1, "Entering Runalong at %d,%d\n",xs,ys);
	abort=0; x=xs; y=ys; n=0; xp=-1; yp=-1;
	while (!abort)
	{

		/* Determine if we got lost in a maze of twisty passages, all alike */

		buf = ((short*)img->data)[x + img->xmax*y];
		if ((buf==0) || (buf==5) || (buf==2))
		{
			dp (1, "Runalong: Fatal %d at %d,%d\n",buf,x,y);
			return n;
		}

		/* Determine the cost of a step in any of the 8 possible directions */

		k=0;
		for (j=y-1; j<=y+1; j++)
			for (i=x-1; i<=x+1; i++)
			{
				k=0;
				if ((i<0) || (j<0) || (i>=img->xmax) || (j>=img->ymax)) 
					k-=2000;	/* Out of bounds? Punish. */
				else
				{
					buf = ((short*)img->data)[i + img->xmax*j];		/* Planned direction: Is it a valid pixel? */
					if (buf==0) k-=1000;							/* Background pixel? Punish. */
					if (buf>=5) k-=500;								/* already visited? Punish. */
					/* check the neighborhood of the suggested pixel */
					for (j1=j-1; j1<=j+1; j1++)
						for (i1=i-1; i1<=i+1; i1++)
						{
							if ((i1>=0) && (j1>=0) && (i1<img->xmax) && (j1<img->ymax) && (i1!=j1))
							{
								buf = ((short*)img->data)[i1 + img->xmax*j1];
								if (buf==0) k+=22;			/* Background pixel is good */
								if (buf==5) k-=3;			/* Visited neighbor is not good */
							}
						}
					}

				cost[i-x+1][j-y+1]=k;
			}

		/* Give an additional bonus if we keep our direction */

		if ((xp>=0) && (yp>=0))
		{
			i=x-xp; j=y-yp;		/* previous to current */
			cost[i+1][j+1]+=1;
		}
		dp (3, "Cost matrix:  ");
		for (j=0; j<=2; j++)
		{
			for (i=0; i<=2; i++)
				dp (3, "%5d  ",cost[i][j]);
			dp (3, "\n           :  ");
		}
		dp (3, "\n");

		/* Done examining the neighborhood. Now get the highest-valued direction */
	
		k=0;
		for (j=y-1; j<=y+1; j++)
			for (i=x-1; i<=x+1; i++)
			{
				dp (3, "Direction %d,%d, cost %d\n",i,j,cost[i-x+1][j-y+1]);
				if (((i!=x) || (j!=y)) && (k < cost[i-x+1][j-y+1]))
				{
					k=cost[i-x+1][j-y+1];
					i1=i; j1=j;
					dp (3, "   -> updated. New cost %d\n",k);
				}
			}

		if (k<=0)		/* No next pixel */
			abort=1;
		else
		{
			ydata[n++] = ARITHM(x-xc,y-yc);
			((short*)img->data)[x + img->xmax*y] = 5;		/* Mark as visited */
			dp (2, "Runalong: Step taken: %d,%d to %d,%d... ",x,y,i1,j1);
			xp=x; yp=y; x=i1; y=j1;
		}

		k=0; buf=((short*)img->data)[x + img->xmax*y];
		for (j=y-1; j<=y+1; j++)
			for (i=x-1; i<=x+1; i++)
			{
				if (((short*)img->data)[i + img->xmax*j]==5) k++;
			}
		dp (2, "New pixel has %d visited neighbors and central value of %d\n",k,buf);

		
		if ((!abort) && (n>10))			/* Check for proximity to start pixel */
		{
			if (( abs(x-xs)<2) && ( abs(y-ys)<2)) 
			{
				abort=1;
				dp (1, "Runalong: Found starting point\n");
			}
		}

	}

	return n;
}


/* Similar algorithm. This one is described in Sonka M, Hlavac V, Boyle R:
	Image processing, analysis, and machine vision (2nd ed.), 
	PWS Publishing, Pacific Grove, CA, 1999, pp. 142-143.
	Note that this algorithm REQUIRES the top left boundary point
	as starting point. This pixel "has the minimum column value of all pixels
	of that region having the minimum row value".
	NOTE: Works very well for 8-neighborhood, but untested for 4-neighborhoods.
*/

int runalong2 (image_type *img, int xs, int ys, int xc, int yc, float *ydata)
{
int dir;			/* Direction of the previous move. 0=East, counting CCW */
int newdir;			/* Search direction */
int f8ngh, abort, x,y,xp,yp,n,i,j,k, cnt;
short buf;
			/*    E  NE   N  NW   W  SW   S  SE   E */
int xdisp[9] = {  1,  1,  0, -1, -1, -1,  0,  1,  1};	/* x-offsets for direction. 0=East, counting CCW */
int ydisp[9] = {  0, -1, -1, -1,  0,  1,  1,  1,  0};



	f8ngh=1;
	dp (1, "Entering Runalong2 at %d,%d\n",xs,ys);
	abort=0; x=xs; y=ys; n=0; xp=-1; yp=-1;
	if (f8ngh) dir=7; else dir=3;			/* Initialize starting direction */
	while (!abort)
	{
		if (f8ngh) 					/* Selext initial search direction */
		{
			if (dir & 1)			/* odd (diagonal) direction */
				newdir=(dir+6) & 0x07;
			else					/* Even (H/V) direction */
				newdir=(dir+7) & 0x07;
		}
		else
			newdir=(dir+3) & 0x03;

		/* Search in direction newdir, going CCW, for a valid pixel */

		cnt=0;
		do
		{
			if (f8ngh)
			{
				xp=x+xdisp[newdir]; yp=y+ydisp[newdir];	
			}
			else
			{
				xp=x+xdisp[newdir*2]; yp=y+ydisp[newdir*2];	
			}
			buf = ((short*)img->data)[xp + img->xmax*yp];
			if (buf==0)
			{
				cnt++; newdir++;
				if (f8ngh) newdir &=0x07; else newdir &=0x03;
			}
		}
		while ((cnt<8) && (buf==0));

		if (cnt>=8)								/* should never happen */
		{
			dp (1, "Runalong failed. Running in circles.\n");
			abort=1; return 0;
		}

		/* Valid boundary pixel found in direction newdir. Update x,y,dir; store radius. */

		dp (2, "Runalong: Step taken: %d,%d to %d,%d... ",x,y,xp,yp);
		x=xp; y=yp; dir=newdir;
		ydata[n++] = ARITHM(x-xc,y-yc);
		((short*)img->data)[x + img->xmax*y] = 5;		/* Mark as visited */

		k=0; buf=((short*)img->data)[x + img->xmax*y];
		for (j=y-1; j<=y+1; j++)
			for (i=x-1; i<=x+1; i++)
			{
				if (((short*)img->data)[i + img->xmax*j]==5) k++;
			}
		dp (2, "New pixel has %d visited neighbors and central value of %d\n",k,buf);

		
		if ((!abort) && (n>10))			/* Check for proximity to start pixel */
		{
			if (( abs(x-xs)<2) && ( abs(y-ys)<2)) 
			{
				abort=1;
				dp (1, "Runalong: Found starting point\n");
			}
		}

	}

	return n;
}



/* Driver for interactive cluster description.
	Uses mainimg and clust_floodfill.
	This entire section assumes 8-connectivity. */


void	describe_cluster (int xp, int yp)
{
clust_id_struct cluster;
image_type tempimg;
int x,y, cx, cy, i,j, xs, ys, n;
int xs2, ys2;		/* Top left boundary point */
double minrad, maxrad, rad, Sx, Sxx, r, fcx, fcy, rx, ry, rz;
double aspect, irreg, compact;
double fourr[20], fouri[20];		/* Fourier descriptors */
double M[4][4], phi[8];					/* Moments */
char buf[256];
char report[16000];
float *xdata, *ydata;


	if (allocate_image (&tempimg, B_SHORT, mainimg.xmax, mainimg.ymax,2)==-1)
		fatal_memerr();

	for (y=0; y<mainimg.ymax; y++)
		for (x=0; x<mainimg.xmax; x++)
		{
			if (!F_ISZERO(readbuf_flt(&mainimg,x,y,cur_slice)))
				((short *)tempimg.data)[x+mainimg.xmax*y]=1;
		}

	cluster.num=0; cluster.srcnum=0;		/* not used here */
	cluster.bx1=cluster.bx2=xp;
	cluster.by1=cluster.by2=yp;				/* Initialize bounding box */
	cluster.Sx=0; cluster.Sy=0;
	cluster.size=0;
	cluster.boundsize=0;
	cluster.compactness = cluster.aspect = cluster.irreg = 0.0;
	clust_flood_breakflg=0; clust_flood_recursion_level=0;

	dp (2, "Flood-filling cluster at %d,%d\n",xp,yp);
	clust_floodfill (&tempimg, &cluster, xp, yp, 1);

	if (clust_flood_breakflg)
	{
		ShowMessage ("Error: Cluster floodfill overflow.\nNot a binary image?");
		return;
	}

	if (cluster.size==0)
	{
		sprintf (buf, "Error: No cluster at %d,%d",xp,yp);
		ShowMessage (buf);
		return;
	}

	/* For the second runalong algorithm, we need the top left corner pixel */

	xs=0; ys=0; 			/* silence a compiler warning */
	xs2=-1; ys2=-1;
	y=cluster.by1;
	while ((y<=cluster.by2) && (ys2<0))
	{
		x=cluster.bx1;
		while ((x<=cluster.bx2) && (xs2<0))
		{
			if (((short *)tempimg.data)[x+mainimg.xmax*y]>1)	/* then it's a pixel in our cluster */
			{
				xs2=x; ys2=y;
				dp (2, "Cluster's top-left starting point at %d,%d\n",xs2,ys2);
			}
			x++;
		}
		y++;
	}

	/* Compute the cluster*s centroid */

	fcx = (double)cluster.Sx/(double)cluster.size;
	fcy = (double)cluster.Sy/(double)cluster.size;
	cx = IROUND(cluster.Sx/(float)cluster.size);
	cy = IROUND(cluster.Sy/(float)cluster.size);
	dp (2, "Found a cluster of size %d with centroid %8.2f,%8.2f\n",cluster.size, fcx, fcy);
	sprintf (report, "Shape descriptor for cluster at %d,%d\nCentroid is %d,%d; size %d pixels, %d boundary pixels\n",xp,yp,cx,cy, cluster.size, cluster.boundsize);

	/* At this time, we have a segmented image. Any pixel that belongs to the cluster has a value > 1,
		and any boundary pixel has a value >2. We also know the number of boundary pixels.
		Use the same loop to copy the boundary pixels (ONLY) to slice 2  */

	rad=0; minrad=-1; maxrad=-1; Sx=0; Sxx=0; rz=0;
	for (i=0; i<20; i++) { fourr[i]=0.0; fouri[i]=0.0; }
	for (i=0; i<=3; i++)
		for (j=0; j<=3; j++)
			M[i][j]=0;

	for (y=cluster.by1; y<=cluster.by2; y++)
		for (x=cluster.bx1; x<=cluster.bx2; x++)
		{
			if ( ((short*)tempimg.data)[x+tempimg.xmax*y] == 3)		/* if this is a boundary pixel */
			{
				rad = ARITHM(x-cx,y-cy);							/* Compute vector from centroid */
				if (minrad==-1) minrad=rad; else minrad = MIN(minrad,rad);
				if (maxrad==-1) 
					{ maxrad=rad; xs=x; ys=y; }
				else if (rad>maxrad) 
				{
					maxrad = rad; xs=x; ys=y;
				}
					
				Sx += sqrt(rad); Sxx += rad;
				((short*)tempimg.data)[x+tempimg.xmax*(y+tempimg.ymax)] = 1;	/* create the boundary */
			}

			if ( ((short*)tempimg.data)[x+tempimg.xmax*y] >= 2)		/* if this is a feature pixel */
			{
				rz += (x-fcx)*(y-fcy);		/* Shouldn't this be zero ??? */
				rx = 1; 					/* Note. Avoid the pow function here. Rather, rely on multiplications */
				for (i=0; i<=3; i++)
				{
					ry=1;
					for (j=0; j<=3; j++)
					{
						M[i][j] += rx*ry;
						ry *= y-fcy;
					}
					rx *= x-fcx;
				}
			}

		}

	dp (2, "Cluster: Radii %5.1f and %5.1f. %d boundary pixels\n",sqrt(maxrad), sqrt(minrad), cluster.boundsize);
	dp (2,  "found rz=%f. Starting point for boundary loop is %d,%d\n",rz,xs,ys);

	compact = SQR((float)cluster.boundsize)/(float)cluster.size;
	if (minrad==0) minrad=1;
	aspect = sqrt(maxrad/minrad);
	irreg = SDEV(Sx, Sxx, cluster.boundsize) * cluster.boundsize / Sx;

	sprintf (buf, "Aspect: %6.2f, Compactness: %6.2f, Irregularity: %6.2f%%\n\n", aspect, compact, 100.0*irreg);
	strncat (report, buf, 15000);

	/* Evaluate the moments */

	r=1;	
	for (i=0; i<=6; i++)		/* Again, we avoid using the pow function */
	{
		phi[i]=r;
		r *= M[0][0];
	}

	strncat (report, "Matrix of statistical moments:\n(normalized except [0,0], [1,0], and [0,1]\n", 15000);
	for (i=0; i<=3; i++)
	{
		for (j=0; j<=3; j++)
		{
			if (i+j >= 2) M[i][j] /= phi[ (i+j+2)>>1 ];		/* Normalization */
			sprintf (buf, "   %8.3f  ",M[i][j]);
			strncat (report, buf, 15000);
		}
		strncat (report, "\n", 15000);
	}

	/* After normalization, we can now compute phi1 through phi7. Store in a separate array to give it more flexibility */

	phi[0]=0;
	phi[1]=M[0][2]+M[2][0];
	phi[2]= SQR(M[0][2]-M[2][0])-4*SQR(M[1][1]);
	phi[3]= SQR(M[3][0]-3*M[1][2]) + SQR(3*M[2][1]-M[0][3]);
	phi[4]= SQR(M[3][0]+M[1][2]) + SQR(M[2][1]+M[0][3]); 
	phi[5]=  (M[3][0]-3*M[1][2])*(M[3][0]+M[1][2])*( SQR(M[3][0]+M[1][2]) - 3*SQR(M[0][3]+M[2][1]))
			+ (3*M[2][1]-M[0][3])*(M[2][1]+M[0][3])*(3*SQR(M[1][2]+M[3][0])-SQR(M[2][1]+M[0][3]));
	phi[6]= (M[2][0]-M[0][2])*(SQR(M[1][2]+M[3][0])-SQR(M[2][1]+M[0][3]))
			+ 4*M[1][1]*(M[3][0]+M[1][2])*(M[0][3]+M[2][1]);
	phi[7]= (3*M[2][1]-M[0][3])*(M[1][2]+M[3][0])*(SQR(M[1][2]+M[3][0])-3*SQR(M[2][1]+M[0][3]))
			+ (3*M[1][2]-M[3][0])*(M[2][1]+M[0][3])*(3*SQR(M[1][2]+M[3][0])-SQR(M[2][1]+M[0][3]));

	strncat (report, "Moment invariants Phi1 to Phi7:\n", 15000);
	for (i=1; i<=7; i++)
	{
		sprintf (buf, "Phi%d= %8.3f\n", i,phi[i]); 	strncat (report, buf, 15000);
	}


	/* Prepare to unwrap the circumference. */
	/* Fill holes to make sure there is no 8-connected background "island" */

	xdata = calloc (cluster.boundsize+10, sizeof (float));
	ydata = calloc (cluster.boundsize+10, sizeof (float));

	fill_holes (&tempimg, 0);

	/* Run along the circumference, fill ydata. We have two algorithms to choose from. */

	/*  n=runalong (&tempimg, xs, ys, cx, cy, ydata); */
	n = runalong2 (&tempimg, xs2, ys2, cx, cy, ydata);

	dp (1, "Runalong finished after %d steps\n",n);

	for (i=0; i<n; i++) 
	{
		xdata[i]=i; 
		ydata[i] = sqrt(ydata[i]);
	}
	graphme (xdata, ydata, n);

	/* Compute the Fourier descriptors */
	
	for (i=0; i<20; i++)					/* Only the first 20 coefficients */
	{
		fourr[i]=0; fouri[0]=0;
		for (j=0; j<n; j++)					/* Run-of-the-mill inefficient DFT, but very transparent. */
		{									/* Might use fftw here as well */
			fourr[i] += ydata[j]*cos(-TWOPI*i*j/(float)n)/n;
			fouri[i] += ydata[j]*sin(-TWOPI*i*j/(float)n)/n;
		}
	}

	strncat (report, "\nFourier coefficients:\n",15000);
	for (i=0; i<10; i++)					/* Only the first 10 coefficients */
	{
		r = sqrt(ARITHM(fourr[i],fouri[i]));
		if (i>0) r = ARITHM(fourr[i],fouri[i])/r;
		sprintf (buf, "C%d = %f (%f,%f)\n",i,r,fourr[i],fouri[i]);	strncat (report, buf, 15000);
	}

	r=0; rz=0;
	for (i=1; i<20; i++)
	{
		r+=sqrt(ARITHM(fourr[i],fouri[i]))/i;
		rz+=sqrt(ARITHM(fourr[i],fouri[i]));
	}
	sprintf (buf, "FF = %f\n",r/rz);
	strncat (report, buf, 15000);

	ShowLongMessage (report, "Cluster shape");
	make_minmax (&tempimg);
	secondary_display (&tempimg, 0, "Detected cluster");
	freebuf (&tempimg);
	free (xdata); free (ydata); 


}


/***************************************************************/


