
/*
	measure.c - General measurement functions 
*/

/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "dlgbox.h"
#include "filter.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "morphop.h"
#include "measure.h"
#include "roi.h"
#include "snakes.h"
#include "secondwnd.h"
#include "segment.h"



/***********************************************************************/



/*
	Calibration functions: Depending on how we calibrated the image, 
	return a calibrated value
*/


double calibrate_val (double x)
{
double f;
	if (wedgecalib.calibrated)	/* Intensity to thickness */
	{
	
		f = fabs(log((wedgecalib.Dmin - wedgecalib.Dmax)/(wedgecalib.Dmin-x)) / wedgecalib.Mu);
		dp (1, "Calibration. IM=%f, I0=%f �=%f. Converted %f to %f\n",
					wedgecalib.Dmax, wedgecalib.Dmin, wedgecalib.Mu,x,f);
	
		return f;
					  
	}
	
	return x;

}



/* Meas_distance:
	This function returns the distance in pixels between the last two clicked
	points (requires two clicks in clickstack).
	This is a GUI-only function that is not connected to the macro interpreter.
*/

void meas_distance (image_type *img)
{
vertex3d start, end;
int dx, dy, dz;
char buf[1024];
float len;

	if (clickstack_last < 2)
	{
		ShowMessage ("You need to click two points first");
		return;
	}

	/* Get the two most recent points that have been clicked */

	start = clickstack[clickstack_last-2];
	end   = clickstack[clickstack_last-1];
	if ((start.x==end.x) && (start.y==end.y) && (start.z==end.z))
	{
		ShowMessage ("Distance is 0");
		return;
	}

	dx = end.x - start.x;
	dy = end.y - start.y;
	dz = end.z - start.z;
	display_clickstack=2;			/* Show the line connecting those 2 points in 2D */

	if (img->zmax<=1)		/* 2D image, don't consider Z */
	{
		len = sqrt (SQR(dx)+SQR(dy));
		sprintf (buf, "Distance in x: %d pixels\nDistance in y: %d pixels\nLength of line: %5.1f pixels\n", dx, dy, len);
	}
	else
	{
		len = sqrt (SQR(dx)+SQR(dy)+SQR(dz));
		sprintf (buf, "Distance in x: %d pixels\nDistance in y: %d pixels\nDistance in z: %d pixels\nLength of line: %5.1f pixels\n", dx, dy, dz, len);
	}
	ShowLongMessage (buf, "Distance Measurement");
}





/* Meas_statistics:
	This function computes average gray value in the ROI and
	displays them over all Z slices 
*/



double SD (double sx, double sxx, long n)
{
	return (SDEV(sx,sxx,n));
}


/* Core of the measurement function. Also used in macro functions */

stat_type meas_core (image_type *img, int z1, int z2, int oorflg)
{
int x,y,z;
int bx1,bx2,by1,by2;	/* Bounding box of ROI */
long idx, cnt, zofs;
double d,dr,db,dg;
stat_type result, oor_result;


	cnt = 0; zofs = img->xmax*img->ymax;
	result.n=0; result.Sx=0; result.Sxx=0;
	result.SxR=0; result.SxG=0; result.SxB=0;
	result.SxxR=0; result.SxxG=0; result.SxxB=0;
	oor_result = result;
	if (z1>=img->zmax) z1=img->zmax-1; else if (z1<0) z1=0;
	if (z2>=img->zmax) z2=img->zmax-1; else if (z2<0) z2=0;

	roi_bbox (img, &bx1,&by1,&bx2,&by2);
	for (y=by1; y<=by2; y++)
		for (x=bx1; x<=bx2; x++)
		{
			if (in_roi (x,y))	/* only once for all Z-slices */
			{

				/* Let's take the switch out of the z loop
				   to increase efficiency by a few %... */
				
				cnt++; /* should be the same for all slices */
				
				idx= (x + img->xmax*y) + img->xmax*img->ymax*z1;
				switch (img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char* p = img->data;
						for (z=z1; z<=z2; z++)
						{
							d = (double) p[idx];
							result.n++;
							result.Sx += d; result.Sxx += d*d;
							if (p[idx])
							{
								oor_result.n++;
								oor_result.Sx += d; oor_result.Sxx += d*d;
							}
							idx += zofs;
						}
						break;
					}
					case B_SHORT:
					{
						short* p = img->data;
						for (z=z1; z<=z2; z++)
						{
							d = (double) p[idx];
							result.n++;
							result.Sx += d;
							result.Sxx += d*d;
							if (p[idx])
							{
								oor_result.n++;
								oor_result.Sx += d; oor_result.Sxx += d*d;
							}
							idx += zofs;
						}
						break;
					}
					case B_FLOAT:
					{
						float* p = img->data;
						for (z=z1; z<=z2; z++)
						{
							d = (double) p[idx];
							result.n++;
							result.Sx += d;
							result.Sxx += d*d;
							if (fabs(p[idx]) > 1e-12)
							{
								oor_result.n++;
								oor_result.Sx += d; oor_result.Sxx += d*d;
							}
							idx += zofs;
						}
						break;
					}
					case B_RGB:
					{
						unsigned char* p = img->data;	/* Note p += 3*idx;*/
						for (z=z1; z<=z2; z++)
						{
							result.n++;
							dr = (double) p[3*idx];
							result.SxR += dr; result.SxxR += dr*dr;
							dg = (double) p[3*idx+1];
							result.SxG += dg; result.SxxG += dg*dg;
							db = (double) p[3*idx+2];
							result.SxB += db; result.SxxB += db*db;

							/* Interpret RGB as non-OOR if either pane is
							   non-zero (at least one color is set) */

							if (p[3*idx] || p[3*idx+1] || p[3*idx+2])
							{
								oor_result.SxR += dr; oor_result.SxxR += dr*dr;
								oor_result.SxG += dg; oor_result.SxxG += dg*dg;
								oor_result.SxB += db; oor_result.SxxB += db*db;
								oor_result.n++;
							}
							idx += 3*zofs;
						}
						break;
					}
					
					
				}
			}
		}
	if (oorflg) 
		return oor_result;
	else
		return result;

}


/* This is the GUI component - works on mainimage */

void meas_statistics()
{
int z,cnt;
stat_type *results, *oor_results;
char msg[65535];
char s[256];

	
	results = (stat_type*)calloc (mainimg.zmax, sizeof (stat_type));
	oor_results = (stat_type*)calloc (mainimg.zmax, sizeof (stat_type));
	if ((!results) || (!oor_results))
	{
		dp (1, "Mem alloc failure\n");
		ShowMessage ("Out of memory for statistics");
		return;
	}

	/* calloc should take care of this, but it is more rigorous to
	   explicitly set all values to 0 */

	for (z=0; z<mainimg.zmax; z++)
	{
		results[z].n=0;
		results[z].Sx=0.0; results[z].Sxx=0.0;
		results[z].SxR=0.0; results[z].SxG=0; results[z].SxB=0.0;
		results[z].SxxR=0.0; results[z].SxxG=0; results[z].SxxB=0.0;
		oor_results[z].n=0;
		oor_results[z].Sx=0.0; oor_results[z].Sxx=0.0;
		oor_results[z].SxR=0.0; oor_results[z].SxG=0; oor_results[z].SxB=0.0;
		oor_results[z].SxxR=0.0; oor_results[z].SxxG=0; oor_results[z].SxxB=0.0;

	}

	/* For improved generality we sacrifice efficiency here.
	   Originally, I had meas_core integrated into this function,
	   so results and oor_results were generated in one sweep.
	   To use meas_core more generally, however, and to keep code length
	   within limits, this is no longer maintained. */

	for (z=0; z<mainimg.zmax; z++)
	{
		results[z] = meas_core (&mainimg, z,z,0);
		oor_results[z] = meas_core (&mainimg, z,z,1);
	}

		
	/* Done with the triple loop, let's prepare the results for display.
	   The region should be the same for all z slices in the first case,
	   so use cnt. */
	
	cnt = results[0].n;
	sprintf (s, "Results of area measurement (%d pixels):\n",cnt);
	strcpy (msg,s);
	if (cnt)
	{
		if (mainimg.imgtype==B_RGB)
			sprintf (s, "Slice\t R \t G \t B \t R \t G \t B\n");
		else
			sprintf (s, "Slice\t Avg \t SD\n");
		strcat (msg,s);
		for (z=0; z<mainimg.zmax; z++)
		{
			if (mainimg.imgtype==B_RGB)
			{
				if (wedgecalib.calibrated)
					sprintf (s, "%d \t %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f\n",
						z,results[z].SxR/cnt, results[z].SxG/cnt, results[z].SxB/cnt,
						SD(results[z].SxR,results[z].SxxR,cnt),
						SD(results[z].SxG,results[z].SxxG,cnt),
						SD(results[z].SxB,results[z].SxxB,cnt),
						calibrate_val(results[z].SxR/cnt),
						calibrate_val(results[z].SxG/cnt),
						calibrate_val(results[z].SxB/cnt)
						);
				else
					sprintf (s, "%d \t %f \t %f \t %f \t %f \t %f \t %f \n",
						z,results[z].SxR/cnt, results[z].SxG/cnt, results[z].SxB/cnt,
						SD(results[z].SxR,results[z].SxxR,cnt),
						SD(results[z].SxG,results[z].SxxG,cnt),
						SD(results[z].SxB,results[z].SxxB,cnt));
			
			}
			else
			{
				if (wedgecalib.calibrated)
					sprintf (s, "%d \t %f \t %f \t %f \n",
						z,results[z].Sx/cnt, 
						SD(results[z].Sx,results[z].Sxx,cnt),
						calibrate_val(results[z].Sx/cnt));
				else
					sprintf (s, "%d \t %f \t %f \n",
						z,results[z].Sx/cnt, SD(results[z].Sx,results[z].Sxx,cnt));
					
			}
			
			strncat (msg,s, 65530);	
		}
	}	

	/* Same thing, but for non-OOR (non-zero pixels) */

	sprintf (s, "\nIf zero is interpreted as out of region:\n");
	strncat (msg,s, 65530);
	if (mainimg.imgtype==B_RGB)
		sprintf (s, "Slice\t Area \t R \t G \t B \t R \t G \t B\n");
	else
		sprintf (s, "Slice\t Area \t Avg \t SD\n");
	strncat (msg,s, 65530);

	for (z=0; z<mainimg.zmax; z++) if (oor_results[z].n)
	{
		cnt = oor_results[z].n;
		if (mainimg.imgtype==B_RGB)
		{
			if (wedgecalib.calibrated)
				sprintf (s, "%d \t %d \t %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f\n",
					z,cnt,oor_results[z].SxR/cnt, oor_results[z].SxG/cnt, oor_results[z].SxB/cnt,
					SD(oor_results[z].SxR,oor_results[z].SxxR,cnt),
					SD(oor_results[z].SxG,oor_results[z].SxxG,cnt),
					SD(oor_results[z].SxB,oor_results[z].SxxB,cnt),
					calibrate_val(oor_results[z].SxR/cnt),
					calibrate_val(oor_results[z].SxG/cnt),
					calibrate_val(oor_results[z].SxB/cnt)
					);
			else
				sprintf (s, "%d \t %d \t %f \t %f \t %f \t %f \t %f \t %f \n",
					z,cnt,oor_results[z].SxR/cnt, oor_results[z].SxG/cnt, oor_results[z].SxB/cnt,
					SD(oor_results[z].SxR,oor_results[z].SxxR,cnt),
					SD(oor_results[z].SxG,oor_results[z].SxxG,cnt),
					SD(oor_results[z].SxB,oor_results[z].SxxB,cnt));

		}
		else
		{
			if (wedgecalib.calibrated)
				sprintf (s, "%d \t %d \t %f \t %f \t %f \n",
					z,cnt, oor_results[z].Sx/cnt, 
					SD(oor_results[z].Sx,results[z].Sxx,cnt),
					calibrate_val(oor_results[z].Sx/cnt));
			else
				sprintf (s, "%d \t %d \t %f \t %f \n",
					z,cnt, oor_results[z].Sx/cnt, SD(oor_results[z].Sx,oor_results[z].Sxx,cnt));

		}

		strncat (msg,s, 65530);	

	}	
	
	free (results);
	free (oor_results);

	ShowLongMessage (msg, "Results");
}



/***********************************************************

	Boundary points
	===============
	
	We assume a segmented image with a set of features
	(background==0 and everything else !=0). We also 
	assume that there is a background pixel to the 
	top left (!). We region-grow a mask from that point 
	which covers all background pixels
	connected with the upper left corner.
	The function returns the number of pixels touching
	the background
	Note: This function complements the "fill hole" function
	and uses a similar algorithm.

************************************************************/



int meas_boundary_pts (image_type* img, int z)
{
image_type mask;
float buf; int cnt, bound;
int x,y,i,j; long idx;
unsigned char* p;
int bx1,bx2,by1,by2;	/* Bounding box of ROI */

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p = mask.data;
	
	/* Create the mask and determine mean value */
	
	cnt=0; idx = z * img->xmax * img->ymax;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,z);
			if (fabs(buf) > 1e-12)
			{
				cnt++;						/* Feature area */
				p[x + mask.xmax*y] = 255;	/* Underneath a feature, the mask is 0xff */
			}
		}
	
	if (cnt==0)	return 0;					/* No features -> exit */
	
	/* Region-grow the mask, setting all background pixels to 30 */
	
	rg_regiongrow (&mask, 0);

	/* Now count. Each pixel of value 255 with a neighbor
	   of value 30 is a boundary pixel. */

	roi_bbox (&mainimg, &bx1,&by1,&bx2,&by2);
	dp (2, "Boundary points: Observing bounding box: %d %d - %d %d\n", bx1,by1,bx2,by2);
	
	bound=0; cnt=0;
	for (y=by1; y<=by2; y++)
		for (x=bx1; x<=bx2; x++)
		{
		
			bound=0;
			if (in_roi (x,y) && (p[x + mask.xmax*y] == 255))
			{
				for (i=MAX(y-1,0); (i<=MIN(y+1,img->ymax-1)) && (bound==0); i++)
					for (j=MAX(x-1,0); (j<=MIN(x+1,img->xmax-1)) && (bound==0); j++)
					{
						if (p[j + i*img->xmax] == 30) bound=1;
					}
			}
			if (bound) cnt++;
		}

	return cnt;
}


void meas_boundary_pts_gui ()
{
int bound;
char s[255];

	bound = meas_boundary_pts (&mainimg, cur_slice);
	
	sprintf (s, "Number of boundary points: %d",bound);
	statusbar_display (s);
}






/*********************************************************

	Run-length histogram
	
	Convert an image (2D or 3D)  into a run-length 
	histogram. This requires binning of the gray values
	and possibly the runs.
	
	The result is a 4-image stack. Each image represents
	one angle (0�, 45�, 90�, 135�). In each image, the
	x axis is the binned gray value and y is the run length.
	The number of runs is coded in the intensity.
	
	The function runlength deletes the original image and
	replaces it with the 4-image stack. Graybins is the
	number of bins for the gray value, and runbins is the
	number of bins for the run length.
	If runbins==0, the image diagonal size is used.
	Zslice is the slice to perform the operation on.

***********************************************************/

	/* Note the 3D behavior of this function. The runs are computed in 2D for ALL slices
		and summed up in the histogram. So the output is always a 4-slice 2D stack,
		and input slices are handled like a sequence of images.
	*/

	/* Note on ROIs. This operation always includes the whole image.
		Honoring the ROI is a TODO item.
		However, a segmented image can be offset (OOR is zero and
		all feature runs are nonzero) and the gray0 row in the
		RL histogram ignored. Just make sure that binning doesn't
		mix the zero row with actual information.
	*/


float runl_scaleg, runl_scalel;		/* Unelegant, but I was in a rush - apologies */


#define GBIN(X,Y,Z) ((int)(graybins*(readbuf_flt(img,X,Y,Z)-min)/(1+max-min)));

void runlength (image_type* img, int graybins, int runbins)
{
image_type run;
int x,y,z,x0,y0,n;
int gbin, rbin, pgbin;
int runcnt,runmax,gmax;
int prerun, pass, pdx,pdy,dx,dy,dir;
int boundaryrun,scaleruns;
float max,min,gbinsize,rbinsize;
short* p;
char s[1024];
int xofs[3][3] = { {1,1,1}, {0,0,1}, {0,0,-1} };
int yofs[3][3] = { {0,-1,1}, {0,1,1}, {0,0,1} };
int plane[3][3]= { {0,3,1}, {0,2,1}, {0,0,3} };

	if (runbins==0)
	{
		runbins = (int) (0.5+sqrt(2.0)*MAX(img->xmax, img->ymax));
		scaleruns=0;		/* Flag that we don't do binning of the runs */
		rbinsize=1.0;
	}
	else
		scaleruns=1;		/* Flag that we place the runs into bins as well */

	imgminmax (img, &min,&max);
	gbinsize = (max-min+1)/graybins;


	/* We need to run over the image to obtain the scaling info (prerun==1)
	   then we run over the image once again to compute the actual 
	   histogram (prerun==0)
	   We get all kinds of compiler warnings here (...might be used
	   uninitialized in this function.). The compiler can't know, so
	   silence it by setting some values.
	*/


	runmax = 0; gmax = 0;x0=0; y0=0; pdx=0; pdy=0; p=NULL; dir=0; gbin=0;
	boundaryrun=1;

	prerun=1;
	while (prerun>=0)
	{
		for (z=0; z<img->zmax; z++)
		{
			/* We start from 3 edges. Left edge (decreasing y)
			   controls runs in X (0�) and x,+y (45�). Upper
			   edge (increasing y) continues x,+y (45�) and runs 
			   in y (90�). Upper edge also starts 135�. Right edge
			   continues 135�. 
			
				Refer to this table for an explanation of the pass and direction counts:
			
				Pass	|	n	|	Start edge	|	Direction
				--------+-------+---------------+--------------
					0	|	0	|	Left		|	  0
					0	|	1	|	Left		|	135 (reverse direction scan)
					0	|	2	|	Left		|	 45
					1	|	1	|	Upper		|	 90
					1	|	2	|	Upper		|	 45
					2	|	2	|	Right		|	135

				Each edge point is the start of a scan line. When we start from the 
				left edge, we place the start points bottom to top. Upper edge, left to
				right. Right edge, top to bottom (this is the direction controlled by PASS.
				From the left edge, we scan 3 directions (the second direction is 135°
				which actually should start from the right edge)
			*/

			for (pass=0; pass<3; pass++)		/* Scan start edge */
			{

				for (n=pass; n<3; n++)			/* Scan direction */
				{

					if (pass==0) { x0=0; y0=img->ymax-1; pdx=0; pdy=-1; }	/* left edge */
					else if (pass==1) { x0=0; y0=0; pdx=1; pdy=0; }			/* upper edge */
					else if (pass==2) { x0=img->xmax-1; y0=0; pdx=0; pdy=1; } /* right edge */
					dx = xofs[pass][n];
					dy = yofs[pass][n];
					dir= plane[pass][n];
					dp (2, "Runlength: Pass %d, direction %d. Start (%d,%d) dir (%d,%d) subdir (%d,%d) Plane %d\n",
						pass,n,x0,y0,pdx,pdy,dx,dy, plane[pass][n]);


					/* At this time, we have set the start corner, the edge from which to scan, and the direction.
						So now walk along the selected edge to cover all start points */

					while ((x0>=0) && (x0<img->xmax) && (y0>=0) && (y0<img->ymax))	/* generic boundary */
					{
						x=x0; y=y0;		/* Start point of current run */

						pgbin = GBIN(x,y,z);
						runcnt=0; boundaryrun=1;
						
						/* Here, run along one scanline that starts at (x0,y0) */
						
						while ((x>=0) && (x<img->xmax) && (y>=0) && (y<img->ymax))	/* generic boundary */
						{

							gbin = GBIN(x,y,z);
							if (gbin != pgbin)		/* Changed gray level so start a new run */
							{
								if (prerun==1)
								{
									if (runcnt>runmax) runmax = runcnt;
									if (gbin > gmax) gmax = gbin;
								}
								else if (boundaryrun==0)			/* Store our current run, but discard runs that touch the image edge */
								{
									if (!scaleruns)
										rbin = runcnt-1;											/* Length, used w/o binning */
									else
										rbin = (int)(runcnt/rbinsize);						/* Binning of the length */
									if ((rbin>=0) && (rbin < runbins))
									{
										p[rbin + run.xmax*pgbin + run.xmax*run.ymax*dir] +=1;
										dp (3, "Run: Start %d,%d. Gray bin %d, runlen %d, run bin %d, dir %d\n",x0,y0,pgbin,runcnt,rbin,dir);
									}
									else
										dp (1, "Run: ERROR, illegal run length %d, bin %d, maximum allowed %d\n",runcnt,rbin, runbins-1);

									if (runcnt==0)
										dp (1, "Run: ERROR, Illegal run length of zero - should not happen!\n",runcnt,runbins-1);
								}

								pgbin=gbin; 
								runcnt=1; 		/* Start at 1 b/c we already have one pixel in this run */
								boundaryrun=0;	/* Once we arrived at this point, it can no longer be a boundary run */
							}
							else
								runcnt++;		/* still within a run, increase length counter */

							x+=dx; y+=dy;
						}


// 						We need to disable this part because we need to discard boundary runs at the end of a run	
// 						if (prerun==1)				/* Scaling run - no real data collected yet */
// 						{
// 							if (runcnt>runmax) runmax = runcnt;
// 							if (gbin > gmax) gmax=gbin;
// 						}
// 						else
// 						{
// 							rbin = (runcnt*runbins)/(runmax+1);
// 							p[rbin + run.xmax*pgbin + run.xmax*run.ymax*dir] +=1;
// 							dp (3, "Run- Start %d,%d. Gray bin %d, runlen %d, run bin %d, dir %d\n",x0,y0,pgbin,runcnt,rbin,dir);
// 						}
// 						runcnt=0;


						x0+=pdx; y0+=pdy;		/* Update start point on edge */
					}
				}
			}

		}	/* end of running over z of the original image */


		if (prerun==1)
		{
			if (scaleruns) rbinsize = (float)(runmax+1)/runbins;
			if (scaleruns && (runbins > runmax)) 	/* If the longest run is smaller than the number of bins... */
			{										/* ...reduce the number of bins to accommodate the longest run */
				dp (1, "Found longest run %d smaller than number of bins %d... adjusting.\n",runmax,runbins);
				runbins = runmax+1;
				scaleruns=0;						/* and set run bin size to 1 */
				rbinsize=1.0;
			}

			dp (1, "Run length scaling info: Max run %d, gray bin size %f, run bin size %f\n",runmax, gbinsize, rbinsize);
			sprintf (s, "Run length scaling info: Run bin size %5.1f, gray bin size %5.1f",rbinsize, gbinsize);
			statusbar_display (s);
			runl_scaleg = gbinsize;
			runl_scalel = rbinsize;

			if (allocate_image (&run, B_SHORT, runbins, graybins, 4)==-1)
				fatal_memerr ();
			p = run.data;
		}
		prerun--;
	}


	/* Return the histo image */
	
	handback (img, &run);

}


/* This function calculates the run legnth metrics from a run length image rlimg.
	The metrics are stored in a 44-element array (to be provided by the caller)
	in the following sequence:

		0	|	SRE		Short run emphasis
		1	|	LRE		Long run emphasis
		2	|	LGRE	Low gray-level emphasis
		3	|	HGRE	High gray-level emphasis
		4	|	GLNU	Gray-level nonuniformity
		5	|	RLNU	Run-length nonuniformity
		6	|	RPC		Run percentage
		7	|	SRLGE	Short-run, low-graylevel emphasis
		8	|	LRHGE	Long-run, high-graylevel amphasis
		9	|	SRHGE	Short-run, high-graylevel emphasis
		10	|	LRLGE	Long-run, low-graylevel emphasis

	The sequence repeats from 11 to 21, 22 to 32, and 33 to 43 for
	45, 90, and 135 degrees.

	See Haidekker: Advanced Biomedical Image Analysis, Chapter 8, for more info */


void runlength_stats (image_type *rlimg, float *result)
{
int g,l,z,i;
int sum[4];
float totals[11];
float scale_l, scale_g, l1,g1, buf, aux;


	if ((rlimg->imgtype!=B_SHORT) || (rlimg->zmax!=4))
		ShowMessage ("Runlength Stats: Probably not a run length histogram");
	if ((runl_scalel==0.0) && (runl_scaleg=0.0))
		ShowMessage ("Runlength Stats: Uninitialized data.");

	/* We need the total counts first to convert from N(g,l) to P(g,l) */
	/* Note that we ignore the row with l==0 */

	sum[0]=sum[1]=sum[2]=sum[3]=0;
	for (z=0; z<rlimg->zmax; z++)
	{
		for (g=0; g<rlimg->ymax; g++)		/* columns are gray bins */
			for (l=1; l<rlimg->xmax; l++)	/* Rows are run lengths */
				sum[z] += ((unsigned short*)rlimg->data)[l+rlimg->xmax*g+z*rlimg->xmax*rlimg->ymax];

		dp (1, "Runlength stats: Normalization for %d degrees is %d\n",45*z, sum[z]);
	}


	/* How do we normalize this? First thought, we use the bin size for g1 and l1,
		but on second thought, we are interested in the histogram's shape, right?
		So wouldn't it be appropriate to let the two axes go from 0 to 1?
		(TODO - needs more exploratory work) */

	//scale_l = runl_scalel/rlimg->xmax;
	//scale_g = runl_scaleg/rlimg->ymax;
	scale_g = 1.0/rlimg->ymax;
	scale_l = 1.0/rlimg->xmax;

	for (z=0; z<rlimg->zmax; z++)
	{
		for (i=0; i<11; i++) totals[i]=0.0;

		/* Accumulating starts here */

		for (g=0; g<rlimg->ymax; g++)		/* columns are gray bins */
			for (l=0; l<rlimg->xmax; l++)	/* Rows are run lengths */
			{
				l1 = (l+1)*scale_l; g1=(g+1)*scale_g;
				buf = ((unsigned short*)rlimg->data)[l+rlimg->xmax*g+z*rlimg->xmax*rlimg->ymax]/(float)sum[z];
				totals[0] += buf/SQR(l1);						/* SRE */
				totals[1] += buf*SQR(l1);						/* LRE */
				totals[2] += buf/SQR(g1);						/* LGRE */
				totals[3] += buf*SQR(g1);						/* HGRE */
				if (buf>0) totals[6] += 1.0/(buf*(l1));			/* RPC */
				totals[7] += buf/SQR(g1)/SQR(l1);				/* SRLGE */
				totals[8] += buf*SQR(g1)*SQR(l1);				/* LRHGE */
				totals[9] += buf*SQR(g1)/SQR(l1);				/* SRHGE */
				totals[10]+= buf*SQR(l1)/SQR(g1);				/* LRLGE */
			}

		/* The row-square and column-square sums GLNU and RLNU have yet to be computed */

		for (g=0; g<rlimg->ymax; g++)		/* columns are gray bins */
		{
			buf = 0.0;
			for (l=0; l<rlimg->xmax; l++)	/* Rows are run lengths */
			{
				buf += ((unsigned short*)rlimg->data)[l+rlimg->xmax*g+z*rlimg->xmax*rlimg->ymax]/(float)sum[z];
			}
			totals[4] += SQR(buf);									/* GLNU */
		}

		for (l=0; l<rlimg->xmax; l++)	/* Rows are run lengths */
		{
			buf = 0.0;
			for (g=0; g<rlimg->ymax; g++)		/* columns are gray bins */
			{
				buf += ((unsigned short*)rlimg->data)[l+rlimg->xmax*g+z*rlimg->xmax*rlimg->ymax]/(float)sum[z];
			}
			totals[5] += SQR(buf);									/* RLNU */
		}

		for (i=0; i<11; i++) result[i+z*11] = totals[i];
	}


}



GtkAdjustment* runl_runbin_adj;
GtkAdjustment* runl_graybin_adj;
image_type* run_img;
int runl_2d3d_flag, runl_report_flag;



void on_runl_cancel_clicked (GtkButton *button, gpointer user_data)
{


	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_runl_ok_clicked (GtkButton *button, gpointer user_data)
{
int graybins, runbins, z;
char rpt[16384], sbuf[256];
float run_stats[46];


	undo_nexus (&mainimg);

	graybins = (int)runl_graybin_adj->value;
	runbins  = (int) runl_runbin_adj->value;

	runlength (run_img, graybins, runbins);

	rpt[0]=0;
	if (runl_report_flag)
	{
		runlength_stats (run_img, run_stats);
		strcpy (rpt, "Run length statistics:\n");

		for (z=0; z<4; z++)
		{
			sprintf (sbuf, "Run angle: %d degrees\n", z*45);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   SRE  =  %f\n", run_stats[z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   LRE  =  %f\n", run_stats[1+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   LGRE =  %f\n", run_stats[2+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   HGRE =  %f\n", run_stats[3+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   GLNU =  %f\n", run_stats[4+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   RLNU =  %f\n", run_stats[5+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   RPC  =  %f\n", run_stats[6+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   SRLGE=  %f\n", run_stats[7+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   LRHGE=  %f\n", run_stats[8+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   SRHGE=  %f\n", run_stats[9+z*11]);
			strcat (rpt, sbuf);
			sprintf (sbuf, "   LRLGE=  %f\n\n", run_stats[10+z*11]);
			strcat (rpt, sbuf);
		}
		ShowLongMessage (rpt, "Run-Length Statistics");
	}
	
	if (run_img==&mainimg)
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}
	
	gtk_widget_destroy (GTK_WIDGET(user_data));
}

void on_runl_checkbutton1_toggled (GtkButton *button, gpointer user_data)
{

	runl_2d3d_flag = (*(GtkToggleButton*)button).active;
	dp (2, "Run length: 2D3D flag set to %d\n",runl_2d3d_flag);

}

void on_runl_checkbutton2_toggled (GtkButton *button, gpointer user_data)
{

	runl_report_flag = (*(GtkToggleButton*)button).active;
	dp (2, "Run length: Report flag set to %d\n",runl_report_flag);

}



void create_runlength_wnd (image_type* img)
{
  GtkWidget *runl_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *runl_label1, *runl_label2, *runl_label3, *runl_label4;
  GtkWidget *runl_graybin, *runl_runbin;
  GtkWidget *runl_2d3d, *runl_report;
  GtkWidget *runl_cancel;
  GtkWidget *runl_ok;
  GtkWidget* hsep1;
  GtkWidget* hsep2;

	run_img = img;
	runl_scaleg=0.0; runl_scalel = 0.0;
	runl_2d3d_flag=0; runl_report_flag=0;

	runl_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (runl_wnd), "runl_wnd", runl_wnd);
	gtk_window_set_title (GTK_WINDOW (runl_wnd), "Run Length Histogram");
	gtk_window_set_modal (GTK_WINDOW (runl_wnd), TRUE);

	vbox1 = vbox_in_window (runl_wnd, "runl_vbox1");
	table1 = MakeTable (runl_wnd, vbox1, "table1", 2,6);

	runl_label1 = MakeLabel (runl_wnd,table1,"Gray bins","runl_label1",0,0);
	runl_graybin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (16, 2, 256, 1, 1, 1));
	runl_graybin = MakeSpinbutton (runl_graybin_adj,runl_wnd,table1,"runl_graybin",1,0);

	runl_label2 = MakeLabel (runl_wnd,table1,"Run bins\n0 for no binning","runl_label2",0,1);
	runl_runbin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (16, 0, 256, 1, 1, 1));
	runl_runbin = MakeSpinbutton (runl_runbin_adj,runl_wnd,table1,"runl_runbin",1,1);

	runl_label3 = MakeLabel (runl_wnd,table1,"3D","runl_label3",0,2);
	runl_2d3d = Checkbutton_in_table (runl_wnd, table1, "2D / 3D","runl_check1",1,2, on_runl_checkbutton1_toggled);

	runl_label4 = MakeLabel (runl_wnd,table1,"Include report","runl_label4",0,3);
	runl_report = Checkbutton_in_table (runl_wnd, table1, " ","runl_check2",1,3, on_runl_checkbutton2_toggled);

	hsep1 = MakeHsep (runl_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (runl_wnd,table1,"hsep2",1,4);
	
	runl_cancel = Button_in_table (runl_wnd,table1,"Cancel","runl_cancel",1,5,NULL);
	runl_ok = Button_in_table (runl_wnd,table1,"OK","runl_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (runl_wnd), "destroy", GTK_SIGNAL_FUNC (on_runl_cancel_clicked), runl_wnd);
	gtk_signal_connect (GTK_OBJECT (runl_ok), "clicked", GTK_SIGNAL_FUNC (on_runl_ok_clicked), runl_wnd);
	gtk_signal_connect (GTK_OBJECT (runl_cancel), "clicked", GTK_SIGNAL_FUNC (on_runl_cancel_clicked), runl_wnd);

	gtk_widget_show (runl_wnd);
}


void runlength_main ()
{

	create_runlength_wnd (&mainimg);

}





