
/*
 *	Segment.c: Algorithms for segmentation
 * (excluding spiderop and snakes - they have extra files)
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "segment.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "dlgbox.h"
#include "application.h"
#include "secondwnd.h"
#include "roi.h"
#include "functions.h"
#include "special1.h"
#include "filter.h"
#include "findclust.h"




/*****************************************************

	Helper functions.

	img_avg determines the average image value 
	inside a box with 2*radius side length centered 
	at position (x,y,z) 

	environ checks if a point at (x,y) may be set.
	This is the case if the point lies within the
	ROI and at least one neighbor is non-zero.
	
	environ3d is the same for 3D. While points in
	the x-y-plane can be 8-connected, adjoining slices
	have to be 2-connected.

******************************************************/


float img_avg (image_type* img, int x,int y,int z,int radius)
{
int i,j;
int cnt;
float avg;

	if (radius==0) return readbuf_flt (img,x,y,z);

	avg=0; cnt=0;
	for (i=x-radius; i<=x+radius; i++)
		for (j=y-radius; j<=y+radius; j++)
			if ((i>=0) && (i<img->xmax) && (j>=0) && (j<img->ymax))
			{
				avg += readbuf_flt (img,i,j,z);
				cnt++;
			}
			
	if (cnt==0) 
		return 0.0;
	else
		return avg/cnt;
}


/* Determine if there is a pixel set in the 3x3 neighborhood of (x,y) */

int environ (image_type* img, int x,int y)
{
int i,j;
int radius=1;

	if (!in_roi(x,y)) return 0;	/* Obey ROI if defined */


	if (img->imgtype==B_CHAR)
	{
		unsigned char *p = img->data;

		for (i=x-radius; i<=x+radius; i++)
			for (j=y-radius; j<=y+radius; j++)
				if ((i>=0) && (i<img->xmax) && (j>=0) && (j<img->ymax))
				{
					if (p[i + img->xmax*j]) return 1;
				}
	}
	else if (img->imgtype==B_SHORT)
	{
		short *p = img->data;

		for (i=x-radius; i<=x+radius; i++)
			for (j=y-radius; j<=y+radius; j++)
				if ((i>=0) && (i<img->xmax) && (j>=0) && (j<img->ymax))
				{
					if (p[i + img->xmax*j]) return 1;
				}
	}


	return 0;
}


int environ3d (image_type* img, int x,int y, int z)
{
int i,j; long idx;
int radius=1;
unsigned char* p;

	if (!in_roi(x,y)) return 0;	/* Obey ROI if defined */

	g_assert (img->imgtype==B_CHAR);
	p = img->data; idx = img->xmax * img->ymax * z;

	for (i=x-radius; i<=x+radius; i++)
		for (j=y-radius; j<=y+radius; j++)
			if ((i>=0) && (i<img->xmax) && (j>=0) && (j<img->ymax))
			{
				if (p[i + img->xmax*j + idx]) return 1;
			}
			
	if ((z>0) && (p[x + img->xmax*(y + img->ymax*(z-1))])) return 1;
	if ((z<img->zmax-1) && (p[x + img->xmax*(y + img->ymax*(z+1))])) return 1;
	return 0;
}


/* Alternative version, no ROI, selectable 2D / 3D. NOTE that 3D is 4-connected only */

int environ2d3d (image_type* img, int x,int y, int z, int is_3d)
{
int i,j,k; long idx;
int radius=1;
unsigned char* p;

	g_assert (img->imgtype==B_CHAR);
	p = img->data; idx = img->xmax * img->ymax * z;

	for (i=x-radius; i<=x+radius; i++)
		for (j=y-radius; j<=y+radius; j++)
			if ((i>=0) && (i<img->xmax) && (j>=0) && (j<img->ymax))
			{
				if (p[i + img->xmax*j + idx]) return 1;
			}
			
	if (!is_3d) return 0;			/* If slected 2D, we are done */
	
	if ((z>0) && (p[x + img->xmax*(y + img->ymax*(z-1))])) return 1;
	if ((z<img->zmax-1) && (p[x + img->xmax*(y + img->ymax*(z+1))])) return 1;
	return 0;
}






/**********************************************************************************************************/

/*         ****    *****    ***     ***     ***    *   *            ***    ****     ***    *   *          */
/*         *   *   *       *   *     *     *   *   *   *           *   *   *   *   *   *   *   *          */
/*         *   *   *       *         *     *   *   **  *           *       *   *   *   *   *   *          */
/*         ****    ****    *  **     *     *   *   * * *           *  **   ****    *   *   * * *          */
/*         * *     *       *   *     *     *   *   *  **           *   *   * *     *   *   * * *          */
/*         *  *    *       *   *     *     *   *   *   *           *   *   *  *    *   *   * * *          */
/*         *   *   *****    ***     ***     ***    *   *            ***    *   *    ***     * *           */
/*                                                                                                        */

/* The post-processing step for the lung segmentation. First,
   flood-fill the background to isolate deselected areas inside
   the lung. then, patch those isolated areas. Finally, erode
   the outside boundary to remove the transition.
   The mask is assumed to be 2D. The mask is a binary B_CHAR
   image and contains the values 0 for background and 255 (!)
   for foreground. The various values are set to allow visualization.
   In the end, anything below 128 is background (zero) and
   anything above is foreground (1) */



/* NOTE: Region growing is incredibly sped up when the scan directions
   alternate per sweep */

void rg_regiongrow (image_type* img, int z)
{
unsigned char* p;
long idx;
int itercnt, changecnt;
int x,y;

	g_assert (img->imgtype==B_CHAR);
	p = img->data;

	itercnt=0; p[img->xmax * img->ymax * z]=30;
	do
	{
		itercnt++; changecnt=0;
		dp (1, "Iteration %d: ", itercnt);

		if (itercnt & 0x01)
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					idx = x + img->xmax* (y + img->ymax*z);
					if (p[idx]==0)
					{
						if (((x>0) && (p[idx-1]==30))
							|| ((x<img->xmax-1) && (p[idx+1]==30))
							|| ((y>0) && (p[idx-img->xmax]==30))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==30)))
						{
							p[idx] = 30;
							changecnt++;
						}
					}
				}
		}
		else
		{
			for (y=img->ymax-1; y>=0; y--)
				for (x=img->xmax-1; x>=0; x--)
				{
					idx = x + img->xmax* (y + img->ymax*z);
					if (p[idx]==0)
					{
						if (((x>0) && (p[idx-1]==30))
							|| ((x<img->xmax-1) && (p[idx+1]==30))
							|| ((y>0) && (p[idx-img->xmax]==30))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==30)))
						{
							p[idx] = 30;
							changecnt++;
						}
					}
				}
		}

	
		dp (1, "%d pixels\n",changecnt);

	}
	while (changecnt>0);


}


void postproc (image_type* mask, int z)
{
int x,y,j, level;
long i,idx;
unsigned char* p;
unsigned char* p1;

	g_assert (mask->imgtype==B_CHAR);
	p = mask->data;
	p1 = (unsigned char*) calloc (mask->xmax*mask->ymax, sizeof (unsigned char));
	if (!p1) fatal_memerr();
	
	level = 0;
	/* rg_floodfill (mask, 1,1,z, &level);	Floodfill the background */
	rg_regiongrow (mask,z);
	
	idx = z*mask->xmax*mask->ymax;
	for (y=0; y<mask->ymax; y++)
		for (x=0; x<mask->xmax; x++)
			if (p[x + mask->xmax*y+idx] == 0)	/* isolated hole */
				p[x + mask->xmax*y+idx] = 200;	/* patch it */
				
	for (j=0; j<2; j++) /* Two iterations of the erosion */
	{
		for (y=1; y<mask->ymax-1; y++)
			for (x=1; x<mask->xmax-1; x++)
			{
				idx = x + mask->xmax*(y+mask->ymax*z);
				if (p[idx] >= 200)	/* lung area */
				{
					if ((p[idx-1]<128) || (p[idx+1]<128) 
						|| (p[idx-mask->xmax]<128) || (p[idx+mask->xmax]<128)
						|| (p[idx-mask->xmax-1]<128) || (p[idx+mask->xmax+1]<128)
						|| (p[idx-mask->xmax+1]<128) || (p[idx+mask->xmax-1]<128))
						p1[x+mask->xmax*y]=100;	/* Erode flag */	
				}
			}

		idx = z*mask->xmax*mask->ymax;
		for (y=1; y<mask->ymax-1; y++)
			for (x=1; x<mask->xmax-1; x++)
			{
				i = x + mask->xmax*y;
				if (p1[i]==100) p[i+idx]=100; 
				p1[i]=0;
			}
	}
	free (p1);

}


/****************************************************+

	3D region growing. This algorithm starts
	at the 3D seed points provided in the seeds array.
	The average gray value computation remains 2D (!)

******************************************************/

/* TODO: for macros, we need to return an error code in case of illegal seed pts
         rather than using ShowMessage and having a garage-door sized memory leak */

void regiongrow_3d (image_type* img, vertex3d* seeds, int seednum, float lthresh, float uthresh, int radius, int lungprocessing)
{
unsigned char *p;	/* for the mask image */
image_type mask;
int i,x,y,z,changecnt,itercnt;
int x1, x2, y1, y2;
long idx;
float buf;
char s[256];
char zero[4] = {0,0,0,0};

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, img->zmax)==-1)
		fatal_memerr ();
		
	p = mask.data;	/* This is calloc'ed above, so it should be all-zero */
	mask.minval.imin=0;  mask.maxval.imax=255;

	/* If a ROI is set, stay within ROI */
	
	roi_bbox (img, &x1,&y1,&x2,&y2);

	for (i=0; i<seednum; i++)
	{
		x = seeds[i].x; y = seeds[i].y; z = seeds[i].z;
		buf = img_avg (img,x,y,z,radius);
		if ((buf<lthresh) || (buf>uthresh) || (!in_roi(x,y)))
		{
			sprintf (s, "Error: Illegal seed point at %d,%d,%d",x,y,z);
			ShowMessage (s);
		}
		else
		{
			p[x + mask.xmax*(y + mask.ymax*z)] = 255;	/* Set the seed points */
			dp (1, "Seed point at %d,%d,%d\n",x,y,z);
		}
	}


	/* Region growing starts here */
	
	itercnt=0;
	do
	{
		itercnt++; changecnt=0;
		dp (1, "Iteration %d: ", itercnt);
	
		/* For speed-up, we switch directions in XY here */

		for (z=0; z<img->zmax; z++)
		{
			if (itercnt & 0x01)
			{
				for (y=y1; y<=y2; y++)
					for (x=x1; x<=x2; x++)
					{
						/* Consider only points which are not yet set and whcih have a 
						   "1" in the mask neighborhood */

						idx = x + mask.xmax* (y + mask.ymax*z);
						if (in_roi(x,y) && (p[idx]==0) && environ3d(&mask,x,y,z))
						{
							buf = img_avg (img,x,y,z,radius);
							if ((buf>=lthresh) && (buf<=uthresh))
							{
								p[idx] = 255;
								changecnt++;
							}
						}
							
					}
			}
			else
			{
				for (y=y2; y>=y1; y--)
					for (x=x2; x>=x1; x--)
					{
						/* Consider only points which are not yet set and which have a 
						   "1" in the mask neighborhood */

						idx = x + mask.xmax* (y + mask.ymax*z);
						if (in_roi(x,y) && (p[idx]==0) && environ3d(&mask,x,y,z))
						{
							buf = img_avg (img,x,y,z,radius);
							if ((buf>=lthresh) && (buf<=uthresh))
							{
								p[idx] = 255;
								changecnt++;
							}
						}
							
					}
			}
		}
	
		dp (1, "%d pixels\n",changecnt);

	}
	while (changecnt>0);

	/* post-processing step. Do this slice by slice */

	if (lungprocessing)
		for (z=0; z<img->zmax; z++)
			postproc (&mask,z);

	if (!macrorun) secondary_display (&mask,cur_slice,"Region Growing Mask");


	/* Perform the actual segmentation. Zero all non-selected points */
	
	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				if (p[x + mask.xmax*(y + mask.ymax*z)]<128)
					writebuf (img, &zero, x,y,z);

	make_minmax (img);
	
	freebuf (&mask);


}


/****************************************************+

	2D region growing. This algorithm stays within
	the slice z and grows from the seed points
	provided in the seeds array. Only x and y coordinates
	of the seed points will be observed.

******************************************************/


void regiongrow_2d (image_type* img, int z, vertex3d* seeds, int seednum, float lthresh, float uthresh, int radius, int lungprocessing)
{
unsigned char *p;	/* for the mask image */
image_type mask;
int i,x,y,changecnt,itercnt;
int x1, x2, y1, y2;
float buf;
char s[256];
char zero[4] = {0,0,0,0};

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
		
	p = mask.data;	/* This is calloc'ed above, so it should be all-zero */
	mask.minval.imin=0;  mask.maxval.imax=255;

	/* If a ROI is set, stay within ROI */
	
	roi_bbox (img, &x1,&y1,&x2,&y2);
	
	for (i=0; i<seednum; i++)
	{
		x = seeds[i].x; y = seeds[i].y; buf = img_avg (img,x,y,z,radius);
		if ((buf<lthresh) || (buf>uthresh) || (!in_roi(x,y)))
		{
			sprintf (s, "Error: Illegal seed point at %d,%d",x,y);
			ShowMessage (s);
		}
		else
		{
			p[x + mask.xmax*y] = 255;	/* Set the seed points */
			dp (1, "Seed point at %d,%d\n",x,y);
		}
	}

	itercnt=0;
	do
	{
		itercnt++; changecnt=0;
		if (!(itercnt & 0x07)) dp (2, "Iteration %d: ", itercnt);
	
		for (y=y1; y<=y2; y++)
			for (x=x1; x<=x2; x++)
			{
				/* Consider only points which are not yet set and which have a 
				   "1" in the mask neighborhood */
			
				if (in_roi(x,y) && (p[x + mask.xmax*y]==0) && environ(&mask,x,y))
				{
					buf = img_avg (img,x,y,z,radius);
					if ((buf>=lthresh) && (buf<=uthresh))
					{
						p[x + mask.xmax*y] = 255;
						changecnt++;
					}
				}
			
			}
	
		if (!(itercnt & 0x07)) dp (2, "%d pixels\n",changecnt);
	}
	while (changecnt>0);

	/* Postprocess if required and display the mask */

	if (lungprocessing) postproc (&mask,0);

	secondary_display (&mask,0,"Region Growing Mask");

	/* Perform the actual segmentation. Zero all non-selected points */
	
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
			if (p[x + mask.xmax*y]<128)
				writebuf (img, &zero, x,y,z);

	make_minmax (img);
	
	freebuf (&mask);

}



/* A related function, which may be of use elsewhere, is
   the hole-filling algorithm. We assume background==0
   and everything else !=0. We also assume that there is a
   background pixel to the top left (!). We region-grow a 
   mask from that point which covers all background pixels
   connected with the upper left corner. Any zero regions
   that are left over must be patched and will be filled
   with the histogram mean.
   TODO: This needs to be improved for RGB images.
*/


void fill_holes (image_type* img, int z)
{
image_type mask;
float avg, buf; int cnt;
int x,y; long idx;
unsigned char* p;

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p = mask.data;
	
	/* Create the mask and determine mean value */
	
	avg=0; cnt=0; idx = z * img->xmax * img->ymax;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,z);
			if (fabs(buf) > 1e-12)
			{
				avg+=buf; cnt++;
				p[x + mask.xmax*y] = 255;
			}
		}
	if (cnt>0) avg/=cnt; else avg=1.0;
		
	/* Region-grow the mask, setting all background pixels to 30 */
	
	rg_regiongrow (&mask, 0);
	
	/* Stuff all holes of the original image where the mask is still zero */

	switch (img->imgtype)	
	{
		case B_CHAR:
		{
				unsigned char* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (int)avg;
						}
					}
				break;
		}
		case B_SHORT:
		{
				short* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (int)avg;
						}
					}
				break;
		}
		case B_FLOAT:
		{
				float* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = avg;
						}
					}
				break;
		}
		case B_RGB:
		{
				rgbtriplet* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (rgbtriplet){(int)avg,(int)avg,(int)avg};
						}
					}
				break;
		}
	}

	freebuf (&mask);
}


void fill_holes_main ()
{
int z;
	
	undo_nexus (&mainimg);
	for (z=0; z<mainimg.zmax; z++)
		fill_holes (&mainimg, z);
}


/******************************************
/*                                        */
/*          ***    *   *    ***           */
/*         *   *   *   *     *            */
/*         *       *   *     *            */
/*         *  **   *   *     *            */
/*         *   *   *   *     *            */
/*         *   *   *   *     *            */
/*          ***     ***     ***           */
/*                                        */
/* The region-growing GUI                 */
/*                                        */
/******************************************/


GtkWidget *rg_window;
GtkAdjustment* rg_seedpts_adj;
GtkAdjustment* rg_lthresh_adj;
GtkAdjustment* rg_uthresh_adj;
GtkAdjustment* rg_avgrad_adj;
GtkWidget *rg_2d3d;
GtkWidget *rg_label4;
int lungprocessing;



void  on_rg_2d3d_toggled (GtkWidget *nothing, gpointer user_data)
{

	if (gtk_toggle_button_get_active ((GtkToggleButton*) rg_2d3d))
		gtk_label_set_text (GTK_LABEL(rg_label4), "3D");
	else
		gtk_label_set_text (GTK_LABEL(rg_label4), "2D");

}

void on_rg_ok_clicked ( GtkButton * button, gpointer data )
{
float uthresh, lthresh;
int rad, seed;
int do_3d;
image_type* img;

	img = data;
	undo_nexus (img);

	do_3d = gtk_toggle_button_get_active ((GtkToggleButton*) rg_2d3d);
	lthresh = rg_lthresh_adj->value;
	uthresh = rg_uthresh_adj->value;
	rad =     rg_avgrad_adj->value;
	seed =    rg_seedpts_adj->value;
	
	dp (1, "Region growing: threshold range %f to %f, radius %d, %d seed points\n",
		lthresh, uthresh, rad, seed);

	dp (1, "Clickstack: %d, Seedpoints %d\n", clickstack_last, seed);

	if (clickstack_last<seed)
		ShowMessage ("Not enough seed points");
	else
	{

		if ((do_3d) && (img->zmax>1))
			regiongrow_3d (img, &clickstack[clickstack_last-seed],seed,lthresh,uthresh,rad,lungprocessing);
		else
			regiongrow_2d (img, cur_slice, &clickstack[clickstack_last-seed],seed,lthresh,uthresh,rad,lungprocessing);

		if (img == &mainimg) 
		{
			make_imlib_buffer_from_main_image (cur_slice);
			redraw (1);
		}
	}

	gtk_widget_destroy (rg_window);
}


void on_rg_cancel_clicked ( GtkButton * button, gpointer data )
{


	gtk_widget_destroy (rg_window);
}

void on_rg_toggled ( GtkButton * button, gpointer data )
{

	lungprocessing = !lungprocessing;
}



void create_rg_window (image_type* img)
{
  GtkWidget* vbox;
  GtkWidget *rg_table1;
  GtkWidget *rg_label1;
  GtkWidget *rg_label2;
  GtkWidget *rg_label3;
  GtkWidget *rg_label5;
  GtkWidget *rg_label6;
  GtkWidget *rg_lthresh;
  GtkWidget *rg_uthresh;
  GtkWidget *rg_avgrad;
  GtkWidget *rg_seedpts;
  GtkWidget* rg_toggle;
  GtkWidget *rg_okbutton;
  GtkWidget *rg_cancelbutton;
  
  float min, max;
  
	imgminmax (img, &min, &max);

	rg_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (rg_window), "rg_window", rg_window);
	gtk_container_set_border_width (GTK_CONTAINER (rg_window), 3);
	gtk_window_set_title (GTK_WINDOW (rg_window), "Region Growing");
	vbox = vbox_in_window (rg_window, "rg_vbox");
	rg_table1 = MakeTable (rg_window, vbox, "rg_table1", 2,7);

	rg_label1 = MakeLabel (rg_window,rg_table1,"Lower threshold","rg_label1",0,0);
	rg_label2 = MakeLabel (rg_window,rg_table1,"Upper threshold","rg_label2",0,1);
	rg_label3 = MakeLabel (rg_window,rg_table1,"Averaging radius","rg_label3",0,2);
	rg_label4 = MakeLabel (rg_window,rg_table1,"2D","rg_label4",0,3);
	rg_label5 = MakeLabel (rg_window,rg_table1,"# of seed points","rg_label5",0,4);
	rg_label6 = MakeLabel (rg_window,rg_table1,"Lung post-processing","rg_label6",0,5);

	rg_lthresh_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, min, max, 1, 10, 0));
	rg_lthresh = MakeHScale (rg_lthresh_adj,rg_window,rg_table1,"rg_lthresh",1,0);
	gtk_scale_set_digits (GTK_SCALE (rg_lthresh), 0);

	rg_uthresh_adj = GTK_ADJUSTMENT (gtk_adjustment_new (max, min, max, 1, 10, 0));
	rg_uthresh = MakeHScale (rg_uthresh_adj,rg_window,rg_table1,"rg_uthresh",1,1);
	gtk_scale_set_digits (GTK_SCALE (rg_uthresh), 0);

	rg_avgrad_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 10, 1, 1, 0));
	rg_avgrad = MakeHScale (rg_avgrad_adj,rg_window,rg_table1,"rg_avgrad",1,2);
	gtk_scale_set_digits (GTK_SCALE (rg_avgrad), 0);

	rg_2d3d = Togglebutton_in_table (rg_window,rg_table1,"2D / 3D","rg_2d3d",1,3,on_rg_2d3d_toggled);

	rg_seedpts_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 1, CLICKSTACK_MAX, 1, 1, 0));
	rg_seedpts = MakeSpinbutton (rg_seedpts_adj,rg_window,rg_table1,"rg_seedpts",1,4);

	rg_toggle = Checkbutton_in_table (rg_window, rg_table1, " ", "rg_toggle", 1,5,on_rg_toggled);

	rg_okbutton = Button_in_table(rg_window,rg_table1,"OK","rg_okbutton",0,6,NULL);
	rg_cancelbutton = Button_in_table(rg_window,rg_table1,"Cancel","rg_cancelbutton",1,6,NULL);

	gtk_signal_connect (GTK_OBJECT (rg_window), "destroy",
                      GTK_SIGNAL_FUNC (on_rg_cancel_clicked), rg_window);
	gtk_signal_connect (GTK_OBJECT (rg_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_rg_ok_clicked),img);
	gtk_signal_connect (GTK_OBJECT (rg_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_rg_cancel_clicked),rg_window);

	gtk_widget_show (rg_window);
}



void regiongrow_main ()
{

	lungprocessing=0;
	create_rg_window (&mainimg);
	
}



/**********************************************************************************************************************/

/*   *   *   *    ***    *****   *****   ****    *****    ***    ***    ***       *****   *   *   ****     ***    *   */
/*   *   *   *   *   *     *     *       *   *   *       *   *    *    *   *        *     *   *   *   *   *   *   *   */
/*   *    * *    *         *     *       *   *   *       *        *    *            *     *   *   *   *   *       *   */
/*****     *      ***      *     ****    ****    ****     ***     *     ***         *     *****   ****     ***    *****/
/*   *     *         *     *     *       * *     *           *    *        *        *     *   *   * *         *   *   */
/*   *     *     *   *     *     *       *  *    *       *   *    *    *   *        *     *   *   *  *    *   *   *   */
/*   *     *      ***      *     *****   *   *   *****    ***    ***    ***         *     *   *   *   *    ***    *   */


/*  Hysteresis thresholding thresholds the image at a fairly high threshold (pass 1) and region-grows each 
	first-pass pixel using a lower threshold.
	NOTE - do we want to obey the ROI?
*/


void hysteresis_threshold (image_type *img, float tlow, float thigh)
{
image_type mask;
unsigned char *p;
int x,y,z;
int pixelcnt,itercnt;

	/* Allocate result image. This is a binary mask, so we use B_CHAR */

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, img->zmax)==-1)
		fatal_memerr();
	p = mask.data;
	mask.maxval.imax=1; mask.minval.imin=0;
	dp (1, "Hysteresis threshold: Upper limit %f, lower limit %f\n", thigh, tlow);

	/* Pass 1: Set all mask pixels to 1 where the image value exceeds thigh */
	
	
	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (readbuf_flt (img,x,y,z) >= thigh)
					p[x + img->xmax*(y + img->ymax*z)] = 1;
			}

	/* Pass 2: Iteratively set all mask pixels in the 8-neighborhood of a set
		mask pixel provided that the image value exceeds tlow.
		This specific implementation is short, but slightly on the inefficient side. */

	itercnt = 0;
	do
	{
		pixelcnt  =0;			/* Number of pixels we set during this iteration */
	
		for (z=0; z<img->zmax; z++)
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					if (!p[x + img->xmax*(y + img->ymax*z)]		/* Mask bit must not be set */
						&& (readbuf_flt (img,x,y,z)>=tlow)		/* Do this first for computational efficiency */
						&& environ2d3d (&mask, x,y,z, 0) )	/* Choose 2D for now */
					{
						p[x + img->xmax*(y + img->ymax*z)] = 1;
						pixelcnt++;
					}
				}

		dp (1, "Hyteresis_threshold: Iteration %d, pixels added to mask: %d\n",++itercnt, pixelcnt);	
	}
	while (pixelcnt > 0);

	/* At this time, we can do one of two things.
		- we can hand back the mask, or
		- we can multiply the original image with the mask and return the thresholded image */

	handback (img, &mask);
}


GtkAdjustment *hystthr_minscale_adj;
GtkAdjustment *hystthr_maxscale_adj;


void on_hystthr_minscale_adj_changed (GtkWidget *adj, gpointer user_data)
{
		/* These are here just in case we want to try a preview */
}

void on_hystthr_maxscale_adj_changed (GtkWidget *adj, gpointer user_data)
{

}


void on_hystthr_ok_clicked ( GtkButton * button, gpointer data )
{

	undo_nexus (&mainimg);
	
	hysteresis_threshold (&mainimg, hystthr_minscale_adj->value, hystthr_maxscale_adj->value);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_hystthr_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}




void create_hystthresh_wnd (image_type* img)
{
  GtkWidget *thresh_wnd;
  GtkWidget *hystthr_vbox1;
  GtkWidget *table1;
  GtkWidget *label1, *label2, *hsep1;
  GtkWidget *hystthr_minscale;
  GtkWidget *hystthr_maxscale;
  GtkWidget *hystthr_okbutton;
  GtkWidget *hystthr_cancelbutton;

  float min, max, stepsz;


	imgminmax (img, &min, &max);
	thresh_return.t = min; thresh_return.u = max;
	thresh_return.thresh_type = TH_BINARY;
	stepsz=1.0; 	/* for all int images (rgb, char, short) */
	if (img->imgtype==B_FLOAT) stepsz=0.1;
	
	thresh_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (thresh_wnd), "thresh_wnd", thresh_wnd);
	gtk_window_set_title (GTK_WINDOW (thresh_wnd), "Hysteresis Threshold");
	gtk_window_set_modal (GTK_WINDOW (thresh_wnd), TRUE);

	hystthr_vbox1 = vbox_in_window (thresh_wnd,"hystthr_vbox1");
	table1 = MakeTable (thresh_wnd, hystthr_vbox1, "hy_table1", 2,4);

	label1 = MakeLabel (thresh_wnd,table1,"Upper Threshold","hy_label1",0,0);
	hystthr_maxscale_adj = GTK_ADJUSTMENT (gtk_adjustment_new (min, min, max, stepsz, 10*stepsz, 0));
	hystthr_maxscale = MakeHScale (GTK_ADJUSTMENT(hystthr_maxscale_adj), thresh_wnd, table1, "hystthr_maxscale", 1,0);

	label2 = MakeLabel (thresh_wnd,table1,"Lower Threshold","hy_label2",0,1);
	hystthr_minscale_adj = GTK_ADJUSTMENT (gtk_adjustment_new (max, min, max, stepsz, 10*stepsz, 0));
	hystthr_minscale = MakeHScale (GTK_ADJUSTMENT(hystthr_minscale_adj), thresh_wnd, table1, "hystthr_minscale", 1,1);

	if (img->imgtype!=B_FLOAT)
	{
		gtk_scale_set_digits (GTK_SCALE (hystthr_minscale), 0);
		gtk_scale_set_digits (GTK_SCALE (hystthr_maxscale), 0);
	}
	
	hsep1 = MakeHsep (thresh_wnd, table1, "hsep1", 0,2);
	hsep1 = MakeHsep (thresh_wnd, table1, "hsep2", 1,2);

	hystthr_okbutton = Button_in_table (thresh_wnd,table1,"OK","hystthr_okbutton",0,3,NULL);
	hystthr_cancelbutton = Button_in_table (thresh_wnd,table1,"Cancel","hystthr_cancelbutton",1,3,NULL);

	gtk_signal_connect (GTK_OBJECT (thresh_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_hystthr_cancel_clicked), thresh_wnd);
	gtk_signal_connect (GTK_OBJECT (hystthr_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hystthr_ok_clicked),thresh_wnd);
	gtk_signal_connect (GTK_OBJECT (hystthr_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hystthr_cancel_clicked),thresh_wnd);

	gtk_widget_show (thresh_wnd);
}


void hysteresis_threshold_main (image_type* img)
{

	create_hystthresh_wnd (img);
}



/**************************************************************************************************/

/*          ***    ****     ***    *   *           *   *   *****   ****     ***    *****          */
/*         *   *   *   *   *   *   *   *           ** **   *       *   *   *   *   *              */
/*         *       *   *   *   *   *   *           * * *   *       *   *   *       *              */
/*         *  **   ****    *   *   * * *   *****   * * *   ****    ****    *  **   ****           */
/*         *   *   * *     *   *   * * *           *   *   *       * *     *   *   *              */
/*         *   *   *  *    *   *   * * *           *   *   *       *  *    *   *   *              */
/*          ***    *   *    ***     * *            *   *   *****   *   *    ***    *****          */
/*                                                                                                */
/**************************************************************************************************

	2D grow & merge. This is a modification of the 
	region growing algorithm where multiple regions
	are grown. They merge if their average value
	falls within a threshold range.

	NOTE: This has never been fully tested and may well be incomplete

******************************************************/


#define BINS	50

/* rg_similarity checks for similar regions. Region's stats are defined by
   its average and sdev. Regions are similar if the lower avg plus its SD
   is higher than the higher avg minus its sd.
   Other similarity criteria may be formulated, for example a more rigid
   version where the averages _and_ the SDs have to match independently 
	Increase the variable 'tolerance' and regions are more likely to merge */


int rg_similarity (float tolerance, float Sx1, float Sxx1, int N1, float Sx2, float Sxx2, int N2)
{
float avg1, avg2, sd1, sd2;

	avg1 = Sx1/N1; avg2 = Sx2/N2;
	if (avg1 > avg2)				/* Make sure avg1 is the smaller one */
	{
		avg1 = Sx2/N2; avg2 = Sx1/N1;
		sd1 = SDEV(Sx2,Sxx2,N2); sd2 = SDEV(Sx1,Sxx1,N1);
	}
	else
	{
		sd1 = SDEV(Sx1,Sxx1,N1); sd2 = SDEV(Sx2,Sxx2,N2);
	}

	if (avg1+sd1+tolerance >= avg2-sd2) return 1;		/* Similar */

	return 0;								/* not similar */
}


/* SEED POINT DETERMINATION:
		Several strategies are possible (and intended). Presently, only one strategy is implemented.
		We compute the source image's histogram and check for peaks. In the seed pass, 
		pixels that fall within hte histogram bin of the histogram maximum are set in a fashion
		that connected clusters with the same value (actually, the same bin value range)
		become one seed cluster with a unique number. The purpose of the seed functions is to
		1) enter all seed points into a mask, and
		2) return the number of seed points (or rather seed clusters)

		The present implementation is hopelessly inefficient, but can serve as a starting point
		for further improvements. Note that flood-filling may be a good replacement in this case.

		Also note that grow-merge wants a pre-segmented image.

*/


int growmerge_fill_seed (image_type* img, image_type *mask, int xs, int ys, int z, int n, float binval, float delta)
{
short* p;
long idx;
int itercnt, changecnt, total_set, allow_grow;
int x,y, bx1, bx2, by1, by2;

	g_assert (mask->imgtype==B_SHORT);
	p = mask->data;
	p[xs + img->xmax*ys]=n;					/* Set the initial seed point */

	bx1=MAX(0,xs-2); bx2=MIN(mask->xmax-1, xs+2);
	by1=MAX(0,ys-2); by2=MIN(mask->ymax-1, ys+2);

	itercnt=0; total_set=0;
	do
	{
		itercnt++; changecnt=0;
		dp (3, "        Iteration %d: ", itercnt);

		for (y=by1; y<=by2; y++)
			for (x=bx1; x<=bx2; x++)
			{
				idx = x + img->xmax*y;
				if (p[idx]==0)											/* Check if unvisited */
				{
					allow_grow=0;

					if (((x>0) && (p[idx-1]==n))						/* Check for 4-connected neighbor */
						|| ((x<img->xmax-1) && (p[idx+1]==n))
						|| ((y>0) && (p[idx-img->xmax]==n))
						|| ((y<img->ymax-1) && (p[idx+img->xmax]==n)))
					{
						allow_grow=1;
					}

					if (((x>0) && (p[idx-1]) && (p[idx-1]!=n))						/* Check for 4-connected neighbor that is not n */
						|| ((x<img->xmax-1) && (p[idx+1]) && (p[idx+1]!=n))
						|| ((y>0) && (p[idx-img->xmax]) && (p[idx-img->xmax]!=n))
						|| ((y<img->ymax-1) && (p[idx+img->xmax]) && (p[idx+img->xmax]!=n)))
					{
						allow_grow=0;			/* In that case, we aren't allowed to grow, supersedes the previous. */
					}


					if (allow_grow && (fabs(binval-readbuf_flt(img,x,y,z)) < delta))	/* Check for membership in bin */
					{
						p[idx] = n;									/* Only then add to present cluster */
						if (x<bx1+2) bx1=MAX(0,x-2);
						if (x>bx2-2) bx2=MIN(mask->xmax-1,x+2);
						if (y<by1+2) by1=MAX(0,y-2);
						if (y>by2-2) by2=MIN(mask->ymax-1,y+2);
						changecnt++; total_set++;
					}
				}
			}

		dp (3, " ... %d pixels added\n",changecnt);
	}
	while (changecnt>0);
	return total_set;
}

/* Growmerge_seed sets the seed points in mask. */


int growmerge_seed (image_type *img, image_type *mask, int zslice)
{
short *p;	/* for the mask image */
int i,j, x,y,bin,n;
int x1, x2, y1, y2;
float buf,min,max,delta;
float histo[BINS+5];					/* 50-bin histogram */
float xdata[BINS+5],ydata[BINS+5];

	p = mask->data;

	roi_bbox (img, &x1,&y1,&x2,&y2);	/* If a ROI is set, stay within ROI */
	imgminmax (img, &min, &max);		/* Get image min and max values */
	
	/* In the first step, we need to find seed points. We get those form the 
	   peaks of a filtered histogram */
	
	for (i=0; i<BINS; i++) histo[i]=0;
	delta = (max-min)/BINS;

	for (y=y1; y<=y2; y++)	
		for (x=x1; x<=x2; x++)
		{
			if (in_roi (x,y))
			{
				buf = readbuf_flt (img, x,y,zslice);
				if (!F_ISZERO(buf))					/* zero is not part of any feature */
				{
					bin = (buf-min)/delta;
					histo[bin]+=1;
				}
			}
		}

	/* Now filter the histogram to allow identification of peaks */
	
	for (j=0; j<3; j++)
	{
		ydata[0] = histo[0];
		ydata[1] = 0.25*histo[0]+0.25*histo[2] + 0.5*histo[1];
		for (i=2; i<BINS; i++) ydata[i] = (histo[i-2]+2*histo[i-1]+3*histo[i]+2*histo[i+1]+histo[i+2])/9;
	}
	if (!macrorun)		/* Display the histogram if not within a macro */
	{
		for (i=2; i<BINS; i++) xdata[i]=min+i*delta;
		graphme (xdata, ydata, BINS);
	}

	/* For each identified peak, go over the mask and set the related points. Related means:
		We seek contiguous regions with values inside the bin of the histogram peak.
		Each of these regions gets assigned a unique number and serves as cluster of
		seed points.
	*/
	
	n=1;
	for (i=1; i<BINS-2; i++)
	{
		if (	(ydata[i] > ydata[i+1]) && (ydata[i] > ydata[i-1]) )	/* identified a peak */
		{
			buf = min+(i+0.5)*delta;		/* The corresponding image value */
			dp (1, "Grow-Merge: Histogram peak at %f has first cluster number %d\n",buf,n+1);
			
			for (y=y1; y<=y2; y++)	
				for (x=x1; x<=x2; x++)
				{
					if (in_roi (x,y) 											/* Seed must be witin ROI */
						&& (p[x + mask->xmax*y]==0)								/* ... and be zero, i.e., unvisited */
						&& ( !F_ISZERO(readbuf_flt (img,x,y,zslice)))			/* ... and lie on a non-background pixel */
						&& ( fabs(buf-readbuf_flt(img,x,y,zslice)) < delta))	/* ... and fall within the current bin */
					{
						j=growmerge_fill_seed (img, mask, x, y, zslice, n, buf, delta);		/* ... then set this point as seed point */
						dp (2, "Grow-Merge seed: Grown %d seed points for seed cluster %d from %d, %d\n",j,n,x,y);
						if (j>0) n++;
					}
				}
		}
	}

	return n;		/* Number of seed points */

}


void grow_merge (image_type* img, int zslice, int seedmeth, float tolerance, int clust_sep, int preview)
{
short *p;	/* for the mask image */
image_type mask;
int i,j,k,n,n0,n1, x,y,changecnt,itercnt,regions;
int x1, x2, y1, y2, xk,yk;
float f,buf,min,max;
float *rcnt, *rsx, *rsxx;
int neighbors[9];
float neigh_vals[9];


	if (allocate_image (&mask, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
		
	p = mask.data;	/* This is calloc'ed above, so it should be all-zero */
	mask.minval.imin=0;  mask.maxval.imax=255;

	roi_bbox (img, &x1,&y1,&x2,&y2);	/* If a ROI is set, stay within ROI */
	imgminmax (img, &min, &max);		/* Get image min and max values */

	n=growmerge_seed (img, &mask, zslice);

	/* If we got more than 32000 seed points, we are in trouble. Not only contains mask
		negative values (which later cause a segfault), but also this is an
		unreasonably high number. In that case, just STOP */

	if (n>32000)
	{
		ShowMessage ("Problem: Too many seed points.\nDid you pre-segment the image?");
		return;
	}
	if (preview)
	{
		make_minmax (&mask);
		secondary_display (&mask, 0, "Grow-Merge Preview");
		return;
	}

	/* Allocate the memory for the statistical arrays */
	
	rcnt = calloc (n+10,sizeof (float));
	rsx  = calloc (n+10,sizeof (float));
	rsxx = calloc (n+10,sizeof (float));
	if ((!rcnt) || (!rsx) || (!rsxx)) fatal_memerr ();

	dp (1, "Grow-merge: Starting with %d seed points.\n",n);

	regions=0;
	for (y=y1; y<=y2; y++)					/* Go over the mask again; this time use it to */
		for (x=x1; x<=x2; x++)				/* ...initialize the statistical arrays */
		{
			if ((i=p[x+mask.xmax*y])>0)			/* SIC! don't use == here */
			{
				rcnt[i]++;
				buf = readbuf_flt (img,x,y,zslice);
				rsx[i] += buf;
				rsxx[i] += SQR(buf);
				regions = MAX(i,regions);
			}
		}

	/* At this time, we have a number of regions, contiguously
	   numbered from 1 to 'regions'. For each region, we accumulate
	   statistics as we grow (mean, sdev). */


	itercnt=0;
	do
	{
		itercnt++; changecnt=0;
		if (!(itercnt & 0x07)) dp (2, "Iteration %d: ", itercnt);
	
		for (y=y1; y<=y2; y++)
			for (x=x1; x<=x2; x++)
			{
				/* Consider only points which are not yet set and which have
					one or more nonzero pixels in the mask neighborhood. Any zero-valued pixel
					with a nonzero neighbor is a candidate for growing */
			
				if (in_roi(x,y) && (p[x + mask.xmax*y]==0) && (F_ISZERO( (f=readbuf_flt(img,x,y,zslice) ))))
				{

					/* Find all nonzero neighbors, line'em up in 'neighbors' array */
					
					for (j=0; j<9; j++) neighbors[j]=0;
					k=0;
					for (i=MAX(0,y-1); i<=MIN(mask.ymax-1,y+1); i++)
						for (j=MAX(0,x-1); j<=MIN(mask.xmax-1,x+1); j++)
						{
							if (p[j+mask.xmax*i]) 	/* Note: The central pixel is always zero */
							{
								neighbors[k]=p[j+mask.xmax*i];
								neigh_vals[k] = readbuf_flt (img, j,i,zslice);
								k++;
							}
						}
					
					if (k>0)		/* means, we have at least one nonzero neighbor b/c the central pixel is zero */
					{
						/* Find the region with the gray value closest to the present pixel */
						
						buf = 10*(max-min+1); n=0;		/* Just a large number */
						for (j=0; j<k; j++)
						{
							if ( fabs(f-neigh_vals[j]) < buf)
							{
								buf = fabs(f-neigh_vals[j]);	/* Find the closest neighbor */
								n = j;
							}
						}

						/* Push the neighbor with the lowest grayval distance to position 0 */
						buf = neigh_vals[0]; neigh_vals[0]=neigh_vals[n]; neigh_vals[n]=buf;
						n0 = neighbors[0]; neighbors[0]=neighbors[n]; neighbors[n]=n0;
						
						n0=neighbors[0];
						buf = readbuf_flt (img, x,y,zslice);	/* buf=f? */
						p[x+mask.xmax*y] = n0;	/* Pixel is attached to the neighbor with the nearest grayvalue distance */
						changecnt++;			/* Count the number of added pixels */
						rcnt[n0]++;				/* Update statistics for this cluster */
						rsx[n0] += buf;
						rsxx[n0] += SQR(buf);

						for (j=1; j<k; j++)
							if (neighbors[j] != n0)		/* These are the two regions we can possibly merge */
							{
								n1 = neighbors[j];
								if (rg_similarity (tolerance, rsx[n1],rsxx[n1],rcnt[n1], rsx[n0],rsxx[n0],rcnt[n0]))
								{
									/* The regions numbered n0 and n1 are similar so
										sweep over the image and convert n1 into n0 */
									
									dp (3, "Grow-merge: Merging regions %d and %d\n",n0,n1);
									for (yk=y1; yk<=y2; yk++)
										for (xk=x1; xk<=x2; xk++)
											if (p[xk+mask.xmax*yk]==n1) p[xk+mask.xmax*yk]=n0;

									rcnt[n0] += rcnt[n1]; rsx[n0]+=rsx[n1]; rsxx[n0]+=rsxx[n1]; 
									rcnt[n1]=0; rsx[n1]=0; rsxx[n1]=0;		/* Region n1 no longer exists. remove */
								}

							}
					}

				}
			
			}
	
		if (!(itercnt & 0x07)) dp (2, "%d pixels\n",changecnt);
	}
	while (changecnt>0);


	/* If we requested cluster separation, delete boundary pixels */
	
	if (clust_sep)
	{
		for (y=y1; y<=y2; y++)
			for (x=x1; x<=x2; x++)
				if ((n=p[x+mask.xmax*y]))		/* Check if mask is set. Do not use == here */
				{
					n0=0;
					for (i=MAX(0,y-1); i<=MIN(mask.ymax-1,y+1); i++)
						for (j=MAX(0,x-1); j<=MIN(mask.xmax-1,x+1); j++)
						{
							if ((p[j+mask.xmax*i]) && (n!=p[j+mask.xmax*i])) n0=1;
						}
					if (n0) p[x+mask.xmax*y]=0;		/* Delete identified boundary pixel */
				}

	}


	/* What follows is a clean-up step where the regions are consecutively numbered */

	for (i=0; i<=regions; i++)
	{
		if (rcnt[i]) rsx[i] /= rcnt[i]; else rsx[i]=regions+10;
		rcnt[i]=i;						/* We only need the average, so discard the rest */
	}

	/* Sort regions by ascending average value */

	for (i=0; i<regions; i++)
	{
		buf=rsx[i]; n=i;
		for (j=i+1; j<=regions; j++)		/* Dirtysort. Inefficient, but... */
			if (rsx[j] < buf)
			{
				buf=rsx[j]; n=j;		/* Find minimum. */
			}
		
		if (i!=n)
		{
			buf=rsx[i]; rsx[i]=rsx[n]; rsx[n]=buf;
			buf=rsxx[i]; rsxx[i]=rsxx[n]; rsxx[n]=buf;
			j=rcnt[i]; rcnt[i]=rcnt[n]; rcnt[n]=j;
		}
	}

	/* Translate in mask image */

	for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
			if ((n=p[x+mask.xmax*y]))		/* Do not use == here */
			{
				j=0; while ((j<regions) && (rcnt[j]!=n)) j++;
				if (rcnt[j]==n) p[x+mask.xmax*y]=j;
			}



	free (rcnt); free (rsx); free (rsxx);

	if (preview)
	{
		make_minmax (&mask);
		secondary_display (&mask, 0, "Grow-Merge Preview");
	}
	else
		handback (img, &mask);

}




GtkWidget *gm_window;
GtkAdjustment* gm_tolerance_adj;
GtkWidget *gm_preview;
GtkWidget *gm_bound;


void  on_gm_preview_toggled (GtkWidget *nothing, gpointer user_data)
{


}

void  on_gm_bound_toggled (GtkWidget *nothing, gpointer user_data)
{


}

void on_gm_ok_clicked ( GtkButton * button, gpointer data )
{
int do_preview, do_bound;
float tolerance;
image_type* img;

	img = data;
	undo_nexus (img);

	do_preview = gtk_toggle_button_get_active ((GtkToggleButton*) gm_preview);
	do_bound = gtk_toggle_button_get_active ((GtkToggleButton*) gm_bound);
	tolerance = gm_tolerance_adj->value;
	
	grow_merge (img, cur_slice, 0, tolerance, do_bound, do_preview);

	if (img == &mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}

	gtk_widget_destroy (gm_window);
}


void on_gm_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy (gm_window);
}




void create_gm_window (image_type* img)
{
  GtkWidget* vbox;
  GtkWidget *gm_table1;
  GtkWidget *gm_label1;
  GtkWidget *gm_label2;
  GtkWidget *gm_label3;
  GtkWidget *gm_tolerance;
  GtkWidget *gm_okbutton;
  GtkWidget *gm_cancelbutton;
  
  float min, max;
  
	imgminmax (img, &min, &max);

	gm_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (gm_window), "gm_window", gm_window);
	gtk_container_set_border_width (GTK_CONTAINER (gm_window), 3);
	gtk_window_set_title (GTK_WINDOW (gm_window), "Grow-Merge");
	vbox = vbox_in_window (gm_window, "gm_vbox");
	gm_table1 = MakeTable (gm_window, vbox, "gm_table1", 2,4);

	gm_label1 = MakeLabel (gm_window,gm_table1,"Tolerance","gm_label1",0,0);
	gm_label2 = MakeLabel (gm_window,gm_table1," ","gm_label4",0,1);
	gm_label3 = MakeLabel (gm_window,gm_table1," ","gm_label5",0,2);

	gm_tolerance_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, -500, 500, 1, 10, 0));
	gm_tolerance = MakeHScale (gm_tolerance_adj,gm_window,gm_table1,"gm_lthresh",1,0);
	gtk_scale_set_digits (GTK_SCALE (gm_tolerance), 0);

	gm_preview = Checkbutton_in_table (gm_window,gm_table1,"Preview","gm_prv",1,1,on_gm_preview_toggled);

	gm_bound = Checkbutton_in_table (gm_window, gm_table1, "Separate clusters", "gm_toggle", 1,2,on_gm_bound_toggled);

	gm_okbutton = Button_in_table(gm_window,gm_table1,"OK","gm_okbutton",0,4,NULL);
	gm_cancelbutton = Button_in_table(gm_window,gm_table1,"Cancel","gm_cancelbutton",1,4,NULL);

	gtk_signal_connect (GTK_OBJECT (gm_window), "destroy",
                      GTK_SIGNAL_FUNC (on_gm_cancel_clicked), gm_window);
	gtk_signal_connect (GTK_OBJECT (gm_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_gm_ok_clicked),img);
	gtk_signal_connect (GTK_OBJECT (gm_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_gm_cancel_clicked),gm_window);

	gtk_widget_show (gm_window);
}


void grow_merge_main ()
{
	create_gm_window (&mainimg);
}



/**************************************************************************/

/*         *   *           *   *   *****     *     *   *    ***           */
/*         *  *            ** **   *        * *    *   *   *   *          */
/*         * *             * * *   *       *   *   **  *   *              */
/*         **      *****   * * *   ****    *   *   * * *    ***           */
/*         * *             *   *   *       *****   *  **       *          */
/*         *  *            *   *   *       *   *   *   *   *   *          */
/*         *   *           *   *   *****   *   *   *   *    ***           */
/*                                                                        */
/**************************************************************************

	K-Means Clustering as segmentation method

	k-Means assumes a pre-thresholded image where background is zero.
 	First, we determine the initial centroids using either the Euclidean
 	distance map (UEPs) or local maxima, depending on seedmeth.
*/


typedef struct
{
	int x,y;
	float val,newx,newy;
	int num,cnt;
}
kmlist_type;


float kmdist (int x, int y, float buf, kmlist_type km, char usegrey)
{
float r;

	if (usegrey)
		r = SQR(x - km.x) + SQR(y - km.y) + SQR (buf - km.val);
	else
		r = SQR(x - km.x) + SQR(y - km.y);

	return r;
}

/* Fuzzy c-Means membership function */

float cmdist (int x, int y, kmlist_type km, int dist)
{
float r;

	r = SQR(x - km.x) + SQR(y - km.y);
	if (r>dist) return 0;
	return 1 - r/dist;
}




void k_means_2d (image_type* img, int z, kmlist_type* kmlist, int cnt, int mindist, int maxdist, char usegrey)
{
image_type temp;
int x,y,xmax,ymax,minidx,i,itercnt;
short *ptemp;
float buf, mindst,epsilon;


	if (allocate_image (&temp, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	xmax=img->xmax; ymax=img->ymax;
	ptemp = temp.data;


	/* For each selected pixel, find the closest centroid
	   and assign its number to the pixel (in temp).
	   At the same time, compute the new centroids */
	
	itercnt=0;
	do
	{
		for (i=0; i<cnt; i++)
		{
			kmlist[i].newx = 0.0; kmlist[i].newy = 0.0; kmlist[i].cnt=0;
		}

		/* Assign each pixel to its closest centroid */

		for (y=0; y<ymax; y++)	
			for (x=0; x<xmax; x++)
			{
				buf = readbuf_flt (img,x,y,z);
				if (buf > 0) 					/* Remember: Pre-thresholded */
				{
					mindst = kmdist (x,y,buf, kmlist[0], usegrey);
					minidx = 0;
					for (i=1; i<cnt; i++)	/* Find closest centroid */
					{
						if (kmdist (x,y,buf, kmlist[i], usegrey) < mindst)
						{
							mindst = kmdist (x,y,buf, kmlist[i], usegrey);
							minidx = i;
						}
					}
					ptemp[x + xmax*y] = kmlist[minidx].num;
					kmlist[minidx].newx += x;
					kmlist[minidx].newy += y;
					kmlist[minidx].cnt++;
				}
			}
		
		/* Update the centroids */
		
		epsilon = 0;
		for (i=0; i<cnt; i++)
			if (kmlist[i].cnt>0)
			{
				kmlist[i].newx /= kmlist[i].cnt;
				kmlist[i].newy /= kmlist[i].cnt;
				x = (int)(0.5+kmlist[i].newx);
				y = (int)(0.5+kmlist[i].newy);

				dp (3, "Updating cluster %d from %d,%d to %d,%d\n",i,
					kmlist[i].x,kmlist[i].y, x, y);

				buf = SQR(x-kmlist[i].x) + SQR(y-kmlist[i].y);
				if (buf>epsilon) epsilon = buf;
				kmlist[i].x=x;
				kmlist[i].y=y;
			}
			
		dp (2, "k-Means: Iteration %d, epsilon=%f\n",++itercnt,epsilon);
	}
	while ((epsilon > 1) && (itercnt < 2000));

	handback (img, &temp);

}




void fuzzy_c_means_2d (image_type* img, int z, kmlist_type* kmlist, int cnt, int mindist, int maxdist, char usegrey)
{
image_type temp;
int x,y,xmax,ymax,minidx,i,itercnt;
long idx;
short *ptemp;
float buf, r,w,epsilon,min;
float *U;


	if (allocate_image (&temp, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	xmax=img->xmax; ymax=img->ymax;
	ptemp = temp.data;

	/* As opposed to k-Means, we need the full membership matrix here */
	
	U = calloc (img->xmax * img->ymax * cnt, sizeof (float));
	if (U==NULL) fatal_memerr ();


	/* For each selected pixel, find the closest centroid
	   and assign its number to the pixel (in temp).
	   At the same time, compute the new centroids */
	
	itercnt=0;
	do
	{
		for (i=0; i<cnt; i++)
		{
			kmlist[i].newx = 0.0; kmlist[i].newy = 0.0; kmlist[i].val=0.0;
		}

		/* Compute the membership function for each pixel */

		for (y=0; y<ymax; y++)	
			for (x=0; x<xmax; x++)
			{
				idx = cnt*(x + y*img->xmax);
				buf = readbuf_flt (img,x,y,z);
				if (buf > 0) 					/* Remember: Pre-thresholded */
				{
					w=0.0;
					for (i=0; i<cnt; i++)	/* for each centroid... */
					{
						r = cmdist (x,y,kmlist[i],mindist);
						U[idx+i] = r;
						w += r;
					}

					/* Normalize so that total membership is 1 */
					
					if (w>0)
						for (i=0; i<cnt; i++)	/* for each centroid... */
						{
							U[idx+i] /= w;
							kmlist[i].x += U[idx+i]*x;
							kmlist[i].y += U[idx+i]*y;
							kmlist[i].val += U[idx+i];
						}
					
				}
			}
					
		
		/* Update the centroids */
		
		epsilon = 0;
		for (i=0; i<cnt; i++)
			if (kmlist[i].val>0)
			{
				kmlist[i].newx /= kmlist[i].val;
				kmlist[i].newy /= kmlist[i].val;
				x = (int)(0.5+kmlist[i].newx);
				y = (int)(0.5+kmlist[i].newy);

				dp (3, "Updating cluster %d from %d,%d to %d,%d\n",i,
					kmlist[i].x,kmlist[i].y, x, y);

				buf = SQR(x-kmlist[i].x) + SQR(y-kmlist[i].y);
				if (buf>epsilon) epsilon = buf;
				kmlist[i].x=x;
				kmlist[i].y=y;
			}
			
		dp (2, "c-Means: Iteration %d, epsilon=%f\n",++itercnt,epsilon);
	}
	while ((epsilon > 1) && (itercnt < 2000));

	/* Last, get the maximum membership for each pixel and store in temp */

	for (y=0; y<ymax; y++)	
		for (x=0; x<xmax; x++)
		{
			idx = cnt*(x + y*img->xmax);
			min=U[idx]; minidx=0;

			for (i=1; i<cnt; i++)	/* for each centroid... */
			{
				if (min<U[idx+i]) { min=U[idx+i]; minidx=i; }
			}
			
			ptemp[x + y*img->xmax] = minidx;
		}
			
			


	free (U);

	handback (img, &temp);

}




int k_means_seed (image_type* img, kmlist_type** km_list, int z, int ngh, char seedmeth, int numverts, float thresh)
{
image_type seed;
int e,cnt,x,y,xmax,ymax,i;
unsigned char nullelem[4] = {0,0,0,0};
unsigned char *psrc, *pdest;
cluster_type* clist;
kmlist_type* kmlist;


	if (allocate_image (&seed, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	xmax=img->xmax; ymax=img->ymax;
	e = DATALEN(img->imgtype);
	psrc = img->data; pdest = seed.data;
	memcpy (pdest, psrc + z*xmax*ymax*e, xmax*ymax*e);

	kmlist = calloc (32767, sizeof (kmlist_type));
	if (!kmlist) fatal_memerr();


	/* Part 1: Initialize the seedpoints or initial centroids */

	if (seedmeth==0)			/* clickstack to get seed points */
	{
		cnt = MIN(numverts, clickstack_last);
		for (i=0; i<cnt; i++)
		{
			kmlist[i].x = clickstack[i].x;
			kmlist[i].y = clickstack[i].y;
			kmlist[i].val = readbuf_flt (img,kmlist[i].x,kmlist[i].y,z);
			kmlist[i].num = i+1;
		}
		
	}
	else if (seedmeth==1)		/* UEP as seed method */
	{
		dp (1, "k-Means: Computing UEPs\n");
		thresh_return.t=thresh_return.r=thresh_return.g=thresh_return.b=1;
		cnt=create_UEP (&seed, 0);

		/* Since the UEP allows for ridges and other multiple points,
		   clusterize and use only the centroids */

		clist = calloc (32760, sizeof (cluster_type));
		cnt = findclust_2d (&seed, 0,32700, 1, 1, 0, clist);
		for (i=0; i<cnt; i++)
		{
			kmlist[i].x = clist[i].cog_x;
			kmlist[i].y = clist[i].cog_y;
			kmlist[i].val = readbuf_flt (img,kmlist[i].x,kmlist[i].y,z);
			kmlist[i].num = i+1;
		}
		free (clist);

	}
	else if (seedmeth==2)	/* Local maxima as seed method */
	{
		dp (1, "k-Means: Computing local maxima\n");
		cnt = locmax_transform_2d (&seed, ngh);

		/* Extract all seed points into a table
		   and remove whatever is below threshold */

		cnt=0;
		for (y=0; y<ymax; y++)	
			for (x=0; x<xmax; x++)
			{
				if ((readbuf_flt (&seed,x,y,0) < thresh) || (cnt>=37260))
				{
					writebuf (&seed, &nullelem, x,y,0);
				}
				else
				{
					kmlist[cnt].x = x; kmlist[cnt].y = y;
					kmlist[cnt].val = readbuf_flt (img,x,y,z);
					kmlist[cnt].num=cnt+1;
					cnt++;
				}
			}
	}
	else if (seedmeth==3)	/* Seed a number of points by random */
	{
		cnt = numverts;
		for (i=0; i<cnt; i++)
		{
			kmlist[i].x = rand() % img->xmax;
			kmlist[i].y = rand() % img->ymax;
			kmlist[i].val = readbuf_flt (img,kmlist[i].x,kmlist[i].y,z);
			kmlist[i].num = i+1;
		}
	
	}
	else
		return 0;

	dp (1, "k-Means: %d seed points\n",cnt);
	
	freebuf (&seed);
	*km_list = kmlist;
	return cnt;
}



/****************************************************************
			k-Means GUI
*/


GtkAdjustment* km_thresh_adj;
GtkAdjustment* km_nghsize_adj;
GtkAdjustment* km_mindist_adj;
GtkAdjustment* km_maxdist_adj;
GtkAdjustment* km_clickstack_adj;
GtkWidget *km_toggle;
int km_seedfunction;



void on_km_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy (GTK_WIDGET(data));
}


void on_km_ok_clicked ( GtkButton * button, gpointer data )
{
char seedmeth;
int ngh, numverts, cnt;
float thresh;
int mindist;
int maxdist;
char usegrey;
kmlist_type* kmlist = NULL;

	undo_nexus (&mainimg);

	thresh = km_thresh_adj->value;
	mindist = km_mindist_adj->value;
	maxdist = km_maxdist_adj->value;
	seedmeth = km_seedfunction;
	numverts = km_clickstack_adj->value;
	ngh = km_nghsize_adj->value;
	usegrey = gtk_toggle_button_get_active ((GtkToggleButton*)km_toggle);

	/* Two-step process. First, we create a list of initial vertices (kmlist),
	   then we perform the actual k-Means clustering on the mainimage */

	cnt = k_means_seed (&mainimg, &kmlist, cur_slice, ngh, seedmeth, numverts, thresh);
	k_means_2d (&mainimg, cur_slice, kmlist, cnt, mindist, maxdist, usegrey);
	free (kmlist);

	cur_slice=0;
	make_imlib_buffer_from_main_image (cur_slice);

	gtk_widget_destroy (GTK_WIDGET(data));
}


void km_set_menu (GtkWidget *widget, gpointer user_data)
{
	km_seedfunction = *(int*) user_data;
	if (km_seedfunction==0)
	{
		km_clickstack_adj->upper=5;
		if (km_clickstack_adj->value > 5) km_clickstack_adj->value=5;
	}
	else if (km_seedfunction==3)
	{
		km_clickstack_adj->upper=1000;
	}
}



void create_kmeans_window (image_type* img)
{
  GtkWidget *km_window;
  GtkWidget* vbox;
  GtkWidget *km_table1;
  GtkWidget *km_label1, *km_label2, *km_label3, *km_label4, *km_label5, *km_label6, *km_label7;
  GtkWidget *km_thresh, *km_clickstack;
  GtkWidget *km_nghsize;
  GtkWidget *km_mindist, *km_maxdist;
  GtkWidget *km_optionmenu1, *km_optionmenu1_menu, *glade_menuitem;
  GtkWidget *km_okbutton;
  GtkWidget *km_cancelbutton;
  
  float min, max;
  
	imgminmax (img, &min,&max);

	km_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (km_window), "km_window", km_window);
	gtk_container_set_border_width (GTK_CONTAINER (km_window), 3);
	gtk_window_set_title (GTK_WINDOW (km_window), "k-Means Clustering");
	vbox = vbox_in_window (km_window, "km_vbox");
	km_table1 = MakeTable (km_window, vbox, "km_table1", 2,8);

	km_label1 = MakeLabel (km_window,km_table1,"Seed points by","km_label1",0,0);
	km_label7 = MakeLabel (km_window,km_table1,"Number of clickstack points","km_label7",0,1);
	km_label2 = MakeLabel (km_window,km_table1,"Threshold","km_label2",0,2);
	km_label3 = MakeLabel (km_window,km_table1,"Neighborhood size","km_label3",0,3);
	km_label4 = MakeLabel (km_window,km_table1,"Use img value in distance","km_label4",0,4);
	km_label5 = MakeLabel (km_window,km_table1,"Min allowable distance","km_label5",0,5);
	km_label6 = MakeLabel (km_window,km_table1,"Max allowable distance","km_label6",0,6);

	km_optionmenu1 = Make_Optionmenu (km_window, km_table1,"km_menu",1,0);
	gtk_container_set_border_width (GTK_CONTAINER (km_optionmenu1), 1);
	km_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (km_optionmenu1_menu, GTK_SIGNAL_FUNC(km_set_menu), "Clickstack", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (km_optionmenu1_menu, GTK_SIGNAL_FUNC(km_set_menu), "UEP", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (km_optionmenu1_menu, GTK_SIGNAL_FUNC(km_set_menu), "Local maxima", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (km_optionmenu1_menu, GTK_SIGNAL_FUNC(km_set_menu), "Random", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (km_optionmenu1), km_optionmenu1_menu);
	km_seedfunction=0;

	km_clickstack_adj = GTK_ADJUSTMENT (gtk_adjustment_new (1, 1,5, 1, 1, 0));
	km_clickstack = MakeHScale (km_clickstack_adj,km_window,km_table1,"km_clickstack",1,1);
	gtk_scale_set_digits (GTK_SCALE (km_clickstack), 0);

	km_thresh_adj = GTK_ADJUSTMENT (gtk_adjustment_new (1, min, max, 1, 10, 0));
	km_thresh = MakeHScale (km_thresh_adj,km_window,km_table1,"km_thresh",1,2);
	if (img->imgtype!=B_FLOAT) gtk_scale_set_digits (GTK_SCALE (km_thresh), 0);

	km_nghsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (1, 1, 7, 1,1,0));
	km_nghsize = MakeHScale (km_nghsize_adj,km_window,km_table1,"km_nghsize",1,3);
	gtk_scale_set_digits (GTK_SCALE (km_nghsize), 0);

	km_toggle = Checkbutton_in_table (km_window, km_table1, " ", "km_toggle", 1,4,NULL);

	km_mindist_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, MIN(img->xmax,img->ymax), 1, 10, 0));
	km_mindist = MakeHScale (km_mindist_adj,km_window,km_table1,"km_mindist",1,5);
	gtk_scale_set_digits (GTK_SCALE (km_mindist), 0);

	km_maxdist_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, MIN(img->xmax,img->ymax), 1, 10, 0));
	km_maxdist = MakeHScale (km_maxdist_adj,km_window,km_table1,"km_maxdist",1,6);
	gtk_scale_set_digits (GTK_SCALE (km_maxdist), 0);

	km_okbutton = Button_in_table(km_window,km_table1,"OK","km_okbutton",0,7,NULL);
	km_cancelbutton = Button_in_table(km_window,km_table1,"Cancel","km_cancelbutton",1,7,NULL);

	gtk_signal_connect (GTK_OBJECT (km_window), "destroy",
                      GTK_SIGNAL_FUNC (on_km_cancel_clicked), km_window);
	gtk_signal_connect (GTK_OBJECT (km_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_km_ok_clicked),km_window);
	gtk_signal_connect (GTK_OBJECT (km_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_km_cancel_clicked),km_window);

	gtk_widget_show (km_window);
}



void k_means_main ()
{

	create_kmeans_window (&mainimg);
	
}



/*
TODO: Here is an interesting algorithm, also called ISODATA. Seen at
http://fourier.eng.hmc.edu/e161/lectures/classification/node12.html



The Isodata Algorithm

Isodata stands for Iterative Self-Organizing Data Analysis Techniques. This is a more sophisticated algorithm which allows the number of clusters to be automatically adjusted during the iteration by merging similar clusters and splitting clusters with large standard deviations. We first define the following parameters:

   1. K = number of clusters desired;
   2. I = maximum number of iterations allowed;
   3. P = maximum number of pairs of cluster which can be merged;
   4. $\Theta_N$ = a threshold value for minimum number of samples in each cluster can have (used for discarding clusters);
   5. $\Theta_S$ = a threshold value for standard deviation (used for split operation);
   6. $\Theta_C$ = a threshold value for pairwise distances (used for merge operation).

The algorithm:

    * Step 1. Arbitrarily choose $k$ (not necessarily equal to K) initial cluster centers: $ {\bf m}_1,\;{\bf m}_2,....,{\bf m}_k $ from the data set $\{{\bf x}_i,\;\;i=1,2,...,N\}$.

    * Step 2. Assign each of the $N$ samples to the closest cluster center:

      \begin{displaymath}{\bf x} \in \omega_j\;\;\;\;if\;\;\;D_L({\bf x},{\bf m}_j) =max\;\{\;D_L({\bf x},{\bf m}_i),\;\;\;i=1,\cdots,k) \end{displaymath}

    * Step 3. Discard clusters with fewer than $\Theta_N$ members, i.e., if for any $j$, $ N_j < \Theta_N $, then discard $\omega_j$ and $k \leftarrow k-1$.
    * Step 4. Update each cluster center:

      \begin{displaymath}{\bf m}_j=\frac{1}{N_j} \sum_{{\bf x} \in \omega_j} {\bf x} \;\;\;\;\;(j=1,\cdots,k) \end{displaymath}

    * Step 5. Compute the average distance $D_j$ of samples in cluster $\omega_j$ from their corresponding cluster center:

      \begin{displaymath}D_j=\frac{1}{N_j} \sum_{{\bf x} \in \omega_j} D_L({\bf x},{\bf m}_j) \;\;\;\;\; (j=1,\cdots,k) \end{displaymath}

    * Step 6. Compute the overall average distance of the samples from their respective cluster centers:

      \begin{displaymath} D=\frac{1}{N} \sum_{j=1}^k N_j D_j \end{displaymath}

    * Step 7. If $k \leq K/2$ (too few clusters), go to Step 8; else if $k>2K$ (too many clusters), go to Step 11; else go to Step 14.

      (Steps 8 through 10 are for split operation, Steps 11 through 13 are for merge operation.)
    * Step 8. First step to split. Find the standard deviation vector ${\bf\sigma}_j=[\sigma_1^{(j)},\cdots,\sigma_n^{(j)}]^T$ for each cluster:

      \begin{displaymath}\sigma_i^{(j)} = \sqrt{\frac{1}{N_j} \sum_{{\bf x} \in \omega... ... (x_i-m_i^{(j)})^2}, \;\;\;\;\;(i=1,\cdots,n,\;\;j=1,\cdots,k) \end{displaymath}

      where $m_i^{(j)}$ is the ith component of ${\bf m}_j$ and $\sigma_i$ is the standard deviation of the samples in $\omega_j$ along the ith coordinate axis. $N_j$ is the number of samples in $\omega_j$.

    * Step 9. Find the maximum component of each $\Sigma_j$ and denote it by $\sigma_{max}^{(j)}$; Do this for all $j=1,\cdots,k$.
    * Step 10.

      If for any $\sigma_{max}^{(j)},\;\;\;\;(j=1,\cdots,k)$, all of the following are true
          o $\sigma_{max}^{(j)} > \Theta_S$,
          o $ D_j > D$,
          o $ N_j > 2\Theta_N$

      then split ${\bf m}_j$ into two new cluster centers ${\bf m}_j^+$ and ${\bf m}_j^-$ by adding $\pm \delta$ to the component of ${\bf m}_j$ corresponding to $\sigma_{max}^{(j)}$, where $\delta$ can be $\alpha\;\sigma_{max}^{(j)}$, for some $\alpha>0$. Then delete ${\bf m}_j$ and let $k \leftarrow k+1$. Goto Step 2

      else Go to Step 14.

    * Step 11. First step to merge. Compute the pairwise distances $D_{ij}$ between every two cluster centers:

      \begin{displaymath}D_{ij}=D_L({\bf m}_i,{\bf m}_j),\;\;\;\;\;\mbox{for all $i\neq j$} \end{displaymath}

      and arrange these $k(k-1)/2$ distances in ascending order.

    * Step 12. Find no more than $P$ smallest $D_{ij}$'s which are also smaller than $\Theta_C$ and keep them in ascending order:

      \begin{displaymath}D_{i_1 j_1} \leq D_{i_2 j_2} \leq \cdots \leq D_{i_P j_P} \end{displaymath}

    * Step 13. Perform pairwise merge: for $l=1,\cdots,P$, do the following:

      If neither of ${\bf m}_{i_l}$ nor ${\bf m}_{j_l}$ has been used in this iteration, then merge them to form a new center:

      \begin{displaymath}{\bf m}=\frac{1}{N_{i_l}+N_{j_l}} [ N_{i_l} {\bf m}_{i_l}+N_{j_l}{\bf m}_{j_l}]\end{displaymath}

      Delete ${\bf m}_{i_l}$ and ${\bf m}_{j_l}$, and let $k \leftarrow k-1$.

      Go to Step 2.
    * Step 14. Terminate if maximum number of iterations I is reached. Otherwise go to Step 2.

The Isodata algorithm is more flexible than the K-mean method. But the user has to choose empirically many more parameters listed previously. 

*/


